"""Useful utilities."""
from __future__ import annotations

import errno
import os
import re
import sys
import time
from contextlib import contextmanager
from io import StringIO
from math import inf
from pathlib import Path
from types import TracebackType
from typing import IO, Any, Generator

from myqueue.config import Configuration

try:
    from functools import cached_property
except ImportError:
    def cached_property(method):  # type: ignore
        """Quick'n'dirty implementation of cached_property coming in 3.8."""
        name = f'__{method.__name__}'

        def new_method(self):  # type: ignore
            if not hasattr(self, name):
                setattr(self, name, method(self))
            return getattr(self, name)

        return property(new_method)


def mqhome() -> Path:
    """Don't use "~/" when testing."""
    name = os.getenv('MYQUEUE_TESTING')
    if name:
        return Path(name)
    return Path.home()


@contextmanager
def chdir(folder: Path) -> Generator:
    """Temporarily change directory."""
    dir = os.getcwd()
    os.chdir(str(folder))
    yield
    os.chdir(dir)


def str2number(s: str) -> int:
    """Convert GB, GiB, ...

    >>> str2number('1MiB')
    1048576
    >>> str2number('2GB')
    2000000000
    """
    n = re.split('[MG]', s)[0]
    return int(n) * {'MB': 1_000_000,
                     'GB': 1_000_000_000,
                     'M': 1_000_000,
                     'G': 1_000_000_000,
                     'MiB': 1024**2,
                     'GiB': 1024**3}[s[len(n):]]


def opencew(filename: str) -> IO[bytes] | None:
    """Create and open filename exclusively for writing.

    If master cpu gets exclusive write access to filename, a file
    descriptor is returned (a dummy file descriptor is returned on the
    slaves).  If the master cpu does not get write access, None is
    returned on all processors."""

    try:
        fd = os.open(str(filename), os.O_CREAT | os.O_EXCL | os.O_WRONLY)
    except OSError as ex:
        if ex.errno == errno.EEXIST:
            return None
        raise
    else:
        return os.fdopen(fd, 'wb')


class Lock:
    """File lock."""
    def __init__(self, name: Path, timeout: float = inf):
        self.lock = name
        self.timeout = timeout
        self.locked = False

    def acquire(self) -> None:
        """Wait for lock to become available and then acquire it."""
        t = 0.0
        delta = 0.05
        while True:
            fd = opencew(str(self.lock))
            if fd is not None:
                break
            time.sleep(delta)
            t += delta
            if t > self.timeout:
                raise TimeoutError(self.lock)
            delta *= 2
        self.locked = True

    def release(self) -> None:
        """Release lock."""
        if self.locked:
            self.lock.unlink()
            self.locked = False

    def __enter__(self) -> 'Lock':
        self.acquire()
        return self

    def __exit__(self,
                 type: Exception, value: Exception, tb: TracebackType) -> None:
        self.release()


def plural(n: int, thing: str) -> str:
    """Add "s" to string if needed.

    >>> plural(0, 'hat'), plural(1, 'hat'), plural(2, 'hat')
    ('0 hats', '1 hat', '2 hats')
    """
    if n == 1:
        return '1 ' + thing
    return f'{n} {thing}s'


def is_inside(path1: Path, path2: Path) -> bool:
    """Check if path1 is inside path2.

    >>> is_inside(Path('a/b'), Path('a/'))
    True
    >>> is_inside(Path('a/'), Path('a/b'))
    False
    """
    try:
        path1.relative_to(path2)
    except ValueError:
        return False
    return True


def normalize_folder(folder: Path, root: Path) -> str:
    """Conver folder to string used in SQLite.

    >>> root = Path('/home/user/a/b/')
    >>> normalize_folder(root / 'c', root)
    './c/'
    >>> normalize_folder(root, root)
    './'
    """
    f = str(folder.relative_to(root))
    if f == '.':
        return './'
    return f'./{f}/'


def update_readme_and_completion(test: bool = False) -> None:
    """Update README.rst and commands dict.

    Run this when ever options are changed::

        python3.9+ -m myqueue.utils

    """

    import argparse
    import textwrap

    from myqueue.cli import _main, aliases, commands

    aliases = {command: alias for alias, command in aliases.items()}

    # Path of the complete.py script:
    dir = Path(__file__).parent

    fd = sys.stdout
    sys.stdout = StringIO()

    print('\n.. list-table::')
    print('    :widths: 1 3\n')
    for cmd, (help, description) in commands.items():
        print(f'    * - :ref:`{cmd} <{cmd}>`', end='')
        if cmd in aliases:
            print(f' ({aliases[cmd]})')
        else:
            print()
        print('      -', help.rstrip('.'))

    for cmd, (help, description) in commands.items():
        help = commands[cmd][0].rstrip('.')
        title = f'{cmd.title()}: {help}'
        if cmd in aliases:
            title = title.replace(':', f' ({aliases[cmd]}):')
        print(f'\n\n.. _{cmd}:\n')
        print(f"{title}\n{'-' * len(title)}\n")
        _main(['help', cmd])

    txt = sys.stdout.getvalue()
    txt = txt.replace(':\n\n    ', '::\n\n    ')
    newlines = txt.splitlines()
    sys.stdout = fd

    n = 0
    while n < len(newlines):
        line = newlines[n]
        if line == 'positional arguments:':
            L: list[str] = []
            n += 1
            while True:
                line = newlines.pop(n)
                if not line:
                    break
                if not line.startswith('                '):
                    cmd, _, help = line.strip().partition(' ')
                    L.append(f'{cmd}:\n    {help.strip()}')
                else:
                    L[-1] += ' ' + line.strip()
            newlines[n - 1:n] = L + ['']
            n += len(L)
        n += 1

    if sys.version_info < (3, 10):
        newlines = [line.replace('optional arguments:', 'options:')
                    for line in newlines]

    cli = dir / '..' / 'docs' / 'cli.rst'

    lines = cli.read_text().splitlines()
    a = lines.index('.. computer generated text:')
    if test:
        old = '\n'.join(lines[a + 1:])
        new = '\n'.join(newlines)
        assert new == old
    else:
        lines[a + 1:] = newlines
        cli.write_text('\n'.join(lines) + '\n')

    filename = dir / 'complete.py'

    dct: dict[str, list[str]] = {}

    class MyException(Exception):
        pass

    class Parser:
        def __init__(self, **kwargs: Any):
            pass

        def add_argument(self, *args: str, **kwargs: Any) -> None:
            pass

        def add_subparsers(self, **kwargs: Any) -> 'Parser':
            return self

        def add_parser(self, cmd: str, **kwargs: Any) -> 'Subparser':
            return Subparser(cmd)

        def parse_args(self, args: list[str] = None) -> None:
            raise MyException

    class Subparser:
        def __init__(self, command: str):
            self.command = command
            dct[command] = []

        def add_argument(self, *args: str, **kwargs: Any) -> None:
            dct[self.command].extend(arg for arg in args
                                     if arg.startswith('-'))

    AP = argparse.ArgumentParser
    argparse.ArgumentParser = Parser  # type: ignore
    try:
        _main()
    except MyException:
        pass
    finally:
        argparse.ArgumentParser = AP  # type: ignore

    txt = 'commands = {'
    for command, opts in sorted(dct.items()):
        txt += "\n    '" + command + "':\n        ["
        txt += '\n'.join(textwrap.wrap("'" + "', '".join(opts) + "'],",
                                       width=65,
                                       break_on_hyphens=False,
                                       subsequent_indent='         '))
    txt = txt[:-1] + '}'

    lines = filename.read_text().splitlines()

    a = lines.index('# Beginning of computer generated data:')
    b = lines.index('# End of computer generated data')

    if test:
        assert '\n'.join(lines[a + 1:b]) == txt
    else:
        lines[a + 1:b] = [txt]
        filename.write_text('\n'.join(lines) + '\n')


def convert_done_files() -> None:
    """Convert old done-files to new-style state files."""
    for path in Path().glob('**/*.done'):
        print(path)
        text = path.read_text()
        if text:
            out = f'{{"state": "done",\n "result": {text}}}\n'
        else:
            out = '{"state": "done"}\n'
        path.with_suffix('.state').write_text(out)
        os.unlink(path)


def get_states_of_active_tasks(folder: Path = None) -> dict[str, str]:
    """Get states of active tasks."""
    from myqueue.queue import Queue
    config = Configuration.read(folder)
    with Queue(config, need_lock=False) as queue:
        active = queue.sql(
            'SELECT folder, name, state FROM tasks '
            'WHERE state IN ("q", "h", "r")')
        return {folder + name: state
                for folder, name, state in active}


if __name__ == '__main__':  # pragma: no cover
    assert sys.version_info >= (3, 9)
    os.environ['COLUMNS'] = '80'
    update_readme_and_completion()



"""User configuration (handling of .myqueue/config.py files)."""
from __future__ import annotations

import os
import subprocess
import warnings
from math import inf
from pathlib import Path
from typing import Any


class Configuration:
    def __init__(self,
                 scheduler: str,
                 nodes: list[tuple[str, dict[str, Any]]] = None,
                 parallel_python: str = 'python3',
                 mpiexec: str = 'mpiexec',
                 extra_args: list[str] = None,
                 maximum_total_task_weight: float = inf,
                 default_task_weight: float = 0.0,
                 notifications: dict[str, str] = None,
                 home: Path = None):
        """Configuration object.

        """
        self.scheduler = scheduler
        self.nodes = nodes or []
        self.parallel_python = parallel_python
        self.mpiexec = mpiexec
        self.extra_args = extra_args or []
        self.maximum_total_task_weight = maximum_total_task_weight
        self.default_task_weight = default_task_weight
        self.notifications = notifications or {}
        self.home = home or Path.cwd()
        self.user = os.environ.get('USER', 'root')

    def __repr__(self) -> str:
        args = ', '.join(f'{name.lstrip("_")}={getattr(self, name)!r}'
                         for name in self.__dict__)
        return f'Configuration({args})'

    def __str__(self) -> str:
        lines = []
        for key, value in self.__dict__.items():
            if key == 'nodes':
                lines.append('nodes')
                n = max((len(name) for name, _ in value), default=0)
                for name, dct in value:
                    lines.append(f'  {name:<{n}}  {dct}')
                continue
            lines.append(f'{key:18} {value}')
        return '\n'.join(lines)

    @classmethod
    def read(self, start: Path = None) -> 'Configuration':
        """Find nearest .myqueue/config.py and read it."""
        if start is None:
            start = Path.cwd()
        home = find_home_folder(start)
        config_file = home / '.myqueue' / 'config.py'
        dct: dict[str, dict[str, Any]] = {}
        exec(compile(config_file.read_text(), str(config_file), 'exec'), dct)
        cfg = dct['config']
        if 'scheduler' not in cfg:
            raise ValueError(
                'Please specify type of scheduler in your '
                f'{home}/.myqueue/config.py '
                "file (must be 'slurm', 'lfs', 'pbs' or 'test').  See "
                'https://myqueue.rtfd.io/en/latest/configuration.html')

        if 'mpi' in cfg or 'mpi_implementation' in cfg:
            warnings.warn(
                'The "mpi" and "mpi_implementation" keywords have been '
                f'deprecated. Please remove them from {config_file}')
            cfg.pop('mpi', None)
            cfg.pop('mpi_implementation', None)

        if 'maximum_diskspace' in cfg:
            warnings.warn(
                'The "maximum_diskspace" keyword has been renamed to '
                '"maximum_total_task_weight". '
                f'Please change this in {config_file}')
            cfg['maximum_total_task_weight'] = cfg.pop('maximum_diskspace')

        config = Configuration(**cfg, home=home)
        return config


def find_home_folder(start: Path) -> Path:
    """Find closest .myqueue/ folder."""
    f = start
    while True:
        dir = f / '.myqueue'
        if dir.is_dir():
            return f.absolute().resolve()
        newf = f.parent
        if newf == f:
            break
        f = newf
    raise ValueError('Could not find .myqueue/ folder!')


def guess_scheduler() -> str:
    """Try different scheduler commands to guess the correct scheduler."""
    scheduler_commands = {'sbatch': 'slurm',
                          'bsub': 'lsf',
                          'qsub': 'pbs'}
    commands = []
    for command in scheduler_commands:
        if subprocess.run(['which', command],
                          stdout=subprocess.DEVNULL,
                          stderr=subprocess.DEVNULL).returncode == 0:
            commands.append(command)
    if commands:
        if len(commands) > 1:
            raise ValueError('Please specify a scheduler: ' +
                             ', '.join(scheduler_commands[cmd]
                                       for cmd in commands))
        scheduler = scheduler_commands[commands[0]]
    else:
        scheduler = 'local'
    return scheduler


def guess_configuration(scheduler_name: str = '',
                        queue_name: str = '',
                        in_place: bool = False) -> None:
    """Simple auto-config tool.

    Creates a config.py file.
    """
    from myqueue.schedulers import get_scheduler
    from myqueue.utils import mqhome, str2number

    folder = mqhome() / '.myqueue'
    if not folder.is_dir():
        folder.mkdir()

    name = scheduler_name or guess_scheduler()
    scheduler = get_scheduler(Configuration(scheduler=name))
    nodelist, extra_args = scheduler.get_config(queue_name)
    nodelist.sort(key=lambda ncm: (-ncm[1], str2number(ncm[2])))
    nodelist2: list[tuple[str, int, str]] = []
    done: set[int] = set()
    for name, cores, memory in nodelist:
        if cores not in done:
            nodelist2.insert(len(done), (name, cores, memory))
            done.add(cores)
        else:
            nodelist2.append((name, cores, memory))

    cfg: dict[str, Any] = {'scheduler': scheduler.name}

    if nodelist2:
        cfg['nodes'] = [(name, {'cores': cores, 'memory': memory})
                        for name, cores, memory in nodelist2]
    if extra_args:
        cfg['extra_args'] = extra_args

    text = f'config = {cfg!r}\n'
    text = text.replace('= {', '= {\n    ')
    text = text.replace(", 'nodes'", ",\n    'nodes'")
    text = text.replace(", 'extra_args'", ",\n    'extra_args'")
    text = text.replace('(', '\n        (')
    text = ('# Generated with mq config.\n'
            '#\n'
            '# Please review the list of node types and remove those\n'
            "# that you don't want to use.  Read more about config.py\n"
            '# files here:\n'
            '#\n'
            '#   https://myqueue.readthedocs.io/en/latest/configuration.html\n'
            '\n' +
            text)

    if in_place:
        cfgfile = folder / 'config.py'
        if cfgfile.is_file():
            cfgfile.rename(cfgfile.with_name('config.py.old'))
        cfgfile.write_text(text)
    else:
        print(text)



from __future__ import annotations
import getpass
import smtplib


class SMTP:
    def __init__(self):
        self.emails = []

    def __call__(self, host):
        return self

    def login(self, username, password):
        assert username == 'me'
        assert password == '********'

    def sendmail(self, fro, to, data):
        self.emails.append(data)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def test_notify(mq, monkeypatch):
    smtp = SMTP()
    monkeypatch.setattr(smtplib, 'SMTP_SSL', smtp)
    monkeypatch.setattr(getpass, 'getpass', lambda: '********')
    monkeypatch.setattr('builtins.input', lambda _: 'yes')
    mq('submit math@sin+3.13')
    mq('modify . -s q -E rdA -z')
    mq('modify . -s q -E rdA')
    mq.wait()
    mq('kick')
    test, notification = smtp.emails
    assert 'Subject: Test email from myqueue' in test
    assert 'Subject: MyQueue: 1 done' in notification



from __future__ import annotations
import os
import sys
import threading
import time
from pathlib import Path

import pytest

from myqueue.config import Configuration
from myqueue.schedulers.local import LocalScheduler, Server
from myqueue.submitting import submit_tasks
from myqueue.task import create_task
from myqueue.workflow import collect, run
from myqueue.queue import sort_out_dependencies


@pytest.fixture(scope='function')
def scheduler(tmpdir):
    port = 39998
    dir = os.getcwd()
    home = Path(tmpdir)
    (home / '.myqueue').mkdir()
    config = Configuration('local', home=home)
    os.chdir(tmpdir)
    server = Server(config, port=port)
    thread = threading.Thread(target=server.start)
    thread.start()
    scheduler = LocalScheduler(config)
    import time
    time.sleep(0.5)
    scheduler.port = server.port
    yield scheduler
    scheduler.send('stop')
    thread.join()
    os.chdir(dir)


@pytest.mark.skipif(sys.version_info < (3, 8),
                    reason='requires Python 3.8 or higher')
def test_local_scheduler(scheduler):
    task1 = create_task('shell:sleep+10', tmax='1s')
    i1 = scheduler.submit(task1)
    assert i1 == 1
    task2 = create_task('shell:sleep+5')
    scheduler.submit(task2)
    ids = scheduler.get_ids()
    assert ids == [1, 2]
    scheduler.cancel(2)
    ids = scheduler.get_ids()
    assert ids == [1]


def workflow():
    with run(shell='fail1', name='1'):
        run(shell='echo', name='2')
    with run(shell='fail2', name='3'):
        run(shell='echo', name='4')


@pytest.mark.skipif(sys.version_info < (3, 8),
                    reason='requires Python 3.8 or higher')
def test_local_scheduler2(scheduler):
    tasks = collect(workflow, Path())
    sort_out_dependencies(tasks)
    ids, ex = submit_tasks(scheduler,
                           tasks,
                           verbosity=2,
                           dry_run=False)
    assert len(ids) == 4
    assert ex is None
    for i in range(10):
        if len(scheduler.get_ids()) == 0:
            break
        time.sleep(0.1)
    else:  # no break
        1 / 0
    names = set(path.name[6:]
                for path in Path('.myqueue').glob('local-*-?'))
    assert names == set(['1-2', '3-2', '3-0', '1-0'])



from __future__ import annotations

import os
from datetime import datetime
from math import inf
from pathlib import Path

import numpy as np
import pytest

from myqueue.caching import (CacheFileNotFoundError, decode, encode,
                             json_cached_function)

objects = [
    [27, 1.4, {'a': 7, 'b': [1, 2]}, inf, -inf],
    1 + 2j,
    None,
    datetime.now(),
    np.zeros((0, 3), complex),
    np.zeros((3, 0), complex),
    np.zeros((2, 1), complex),
    np.zeros(2, np.float32),
    np.ones((2, 2), int)]


@pytest.mark.parametrize('obj1', objects)
def test_encoding(obj1):
    """Test encoding/decoding."""
    text1 = encode(obj1)
    obj2 = decode(text1)
    text2 = encode(obj2)
    assert text1 == text2
    print(text1)
    if isinstance(obj1, np.ndarray):
        assert (obj1 == obj2).all()
        assert obj1.shape == obj2.shape
        assert obj1.dtype == obj2.dtype
    else:
        assert obj1 == obj2


def func(a: int, b: int) -> int:
    """Test function."""
    return a + b


def test_no_cache(tmp_path):
    """Test function that returns non-jsonable object."""
    os.chdir(tmp_path)
    function = json_cached_function(func, 'add', [1], {'b': 2})
    with pytest.raises(CacheFileNotFoundError):
        function(only_read_from_cache=True)

    assert function() == 3

    assert Path('add.result').read_text() == '3'
    Path('add.result').write_text('4')
    assert function() == 4



"""Definitions of commands.

There is a Command base class and five concrete classes:
ShellCommand, ShellScript, PythonScript, PythonModule and
PythonFunction.  Use the factory function command() to create
command objects.
"""
from __future__ import annotations
from typing import Any, Type, Callable
from pathlib import Path
from shlex import quote

from myqueue.resources import Resources


class Command:
    """Base class."""
    def __init__(self, name: str, args: list[str]):
        self.args = args
        if args:
            name += '+' + '_'.join(self.args)
        self.name = name
        self.dct: dict[str, Any] = {'args': args}
        self.short_name: str
        self.function: Callable[[], Any] | None = None

    def set_non_standard_name(self, name: str) -> None:
        self.name = name
        self.dct['name'] = name

    def todict(self) -> dict[str, Any]:
        raise NotImplementedError

    @property
    def fname(self) -> str:
        return self.name.replace('/', '\\')  # filename can't contain slashes

    def read_resources(self, path: Path) -> Resources | None:
        """Look for "# MQ: resources=..." comments in script."""
        return None

    def run(self) -> Any:
        import subprocess
        subprocess.run(str(self), shell=True, check=True)
        return None

    def quoted_args(self) -> list[str]:
        return [quote(arg) for arg in self.args]


def create_command(cmd: str,
                   args: list[str] = [],
                   type: str = None,
                   name: str = '') -> Command:
    """Create command object."""
    cmd, _, args2 = cmd.partition(' ')
    if args2:
        args = args2.split() + args
    path, sep, cmd = cmd.rpartition('/')
    if '+' in cmd:
        cmd, _, rest = cmd.rpartition('+')
        args = rest.split('_') + args
    cmd = path + sep + cmd

    cls: Type[Command]
    if type is None:
        if cmd.startswith('shell:'):
            cls = ShellCommand
        elif cmd.endswith('.py'):
            cls = PythonScript
        elif cmd.startswith('workflow:'):
            cls = WorkflowTask
        elif '.py@' in cmd:
            cls = PythonFunctionInScript
        elif '@' in cmd:
            cls = PythonFunction
        elif path:
            cls = ShellScript
        else:
            cls = PythonModule
    else:
        cls = globals()[type.title().replace('-', '')]

    command = cls(cmd, args)

    if name:
        command.set_non_standard_name(name)

    return command


class ShellCommand(Command):
    def __init__(self, cmd: str, args: list[str]):
        Command.__init__(self, cmd, args)
        self.cmd = cmd
        self.short_name = cmd

    def __str__(self) -> str:
        return ' '.join([self.cmd[6:]] + self.quoted_args())

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'shell-command',
                'cmd': self.cmd}


class ShellScript(Command):
    def __init__(self, cmd: str, args: list[str]):
        Command.__init__(self, Path(cmd).name, args)
        self.cmd = cmd
        self.short_name = cmd

    def __str__(self) -> str:
        return ' '.join(['sh', self.cmd] + self.quoted_args())

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'shell-script',
                'cmd': self.cmd}

    def read_resources(self, path: Path) -> Resources | None:
        for line in Path(self.cmd).read_text().splitlines():
            if line.startswith('# MQ: resources='):
                return Resources.from_string(line.split('=', 1)[1])
        return None


class PythonScript(Command):
    def __init__(self, script: str, args: list[str]):
        path = Path(script)
        Command.__init__(self, path.name, args)
        if '/' in script:
            self.script = str(path.absolute())
        else:
            self.script = script
        self.short_name = path.name

    def __str__(self) -> str:
        return 'python3 ' + ' '.join([self.script] + self.quoted_args())

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'python-script',
                'cmd': self.script}

    def read_resources(self, path: Path) -> Resources | None:
        script = Path(self.script)
        if not script.is_absolute():
            script = path / script
        for line in script.read_text().splitlines():
            if line.startswith('# MQ: resources='):
                return Resources.from_string(line.split('=', 1)[1])
        return None


class WorkflowTask(Command):
    def __init__(self,
                 cmd: str,
                 args: list[str],
                 function: Callable[..., Any] = None):
        script, name = cmd.split(':')
        self.script = Path(script)
        Command.__init__(self, name, args)
        self.function = function
        self.short_name = name

    def __str__(self) -> str:
        code = '; '.join(
            ['from myqueue.workflow import run_workflow_function',
             f'run_workflow_function({str(self.script)!r}, {self.name!r})'])
        return f'python3 -c "{code}"'

    def run(self) -> Any:
        assert self.function is not None
        return self.function()

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'workflow-task',
                'cmd': f'{self.script}:{self.name}'}


class PythonModule(Command):
    def __init__(self, mod: str, args: list[str]):
        Command.__init__(self, mod, args)
        self.mod = mod
        self.short_name = mod

    def __str__(self) -> str:
        return ' '.join(['python3', '-m', self.mod] + self.quoted_args())

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'python-module',
                'cmd': self.mod}


class PythonFunction(Command):
    def __init__(self, cmd: str, args: list[str]):
        if ':' in cmd:
            # Backwards compatibility with version 4:
            self.mod, self.func = cmd.rsplit(':', 1)
        else:
            self.mod, self.func = cmd.rsplit('@', 1)
        Command.__init__(self, cmd, args)
        self.short_name = cmd

    def __str__(self) -> str:
        args = ', '.join(repr(convert(arg)) for arg in self.args)
        mod = self.mod
        return f'python3 -c "import {mod}; {mod}.{self.func}({args})"'

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'python-function',
                'cmd': self.mod + '@' + self.func}


class PythonFunctionInScript(Command):
    def __init__(self, cmd: str, args: list[str]):
        script, self.func = cmd.rsplit('@', 1)
        path = Path(script)
        Command.__init__(self, path.name, args)
        if '/' in script:
            self.script = str(path.absolute())
        else:
            self.script = script
        self.short_name = path.name

    def __str__(self) -> str:
        args = ', '.join(repr(convert(arg)) for arg in self.args)
        return (f'python3 -c "import runpy; '
                f'mod = runpy({self.script!r}); '
                f'mod.{self.func}({args})')

    def todict(self) -> dict[str, Any]:
        return {**self.dct,
                'type': 'python-function-in-script',
                'cmd': self.script + '@' + self.func}


def convert(x: str) -> bool | int | float | str:
    """Convert str to bool, int, float or str."""
    if x == 'True':
        return True
    if x == 'False':
        return False
    try:
        f = float(x)
    except ValueError:
        return x
    if int(f) == f:
        return int(f)
    return f



from myqueue.queue import Queue
from myqueue.utils import plural


def sync(queue: Queue) -> None:
    """Syncronize queue with the real world."""
    ids = queue.scheduler.get_ids()
    remove = []

    sql = 'SELECT id FROM tasks WHERE state IN ("q", "h", "r")'
    for id, in queue.sql(sql):
        if id not in ids:
            remove.append(id)

    root = queue.folder.parent
    for id, folder in queue.sql('SELECT id, folder FROM tasks'):
        if not (root / folder).is_dir():
            remove.append(id)

    print('REMOVE', remove)
    if remove:
        if queue.dry_run:
            print(plural(len(remove), 'job'), 'to be removed')
        else:
            queue.remove(remove)
            print(plural(len(remove), 'job'), 'removed')



from __future__ import annotations
from pathlib import Path

from myqueue.states import State
from myqueue.utils import normalize_folder


class Selection:
    """Object used for selecting tasks."""

    def __init__(self,
                 ids: set[int] | None = None,
                 name: str | None = None,
                 states: set[State] = set(),
                 folders: list[Path] = [],
                 recursive: bool = True,
                 error: str | None = None):
        """Selection.

        Selection is based on:

            ids

        or:

            any combination of name, state, folder and error message.

        Use recursive=True to allow for tasks inside a folder.
        """

        self.ids = ids
        self.name = name
        self.states = states
        self.folders = folders
        self.recursive = recursive
        self.error = error

    def __repr__(self) -> str:
        return (f'Selection({self.ids}, {self.name}, {self.states}, '
                f'{self.folders}, {self.recursive}, {self.error})')

    def sql_where_statement(self, root: Path) -> tuple[str, list]:
        if self.ids is not None:
            q = ', '.join('?' * len(self.ids))
            return (f'id IN ({q})', list(self.ids))

        parts = []
        args = []
        if len(self.states) < 8:
            q = ', '.join('?' * len(self.states))
            parts.append(f'state IN ({q})')
            args += [state.value for state in self.states]

        if self.folders:
            part = []
            for folder in self.folders:
                f = normalize_folder(folder, root)
                if self.recursive:
                    part.append('folder GLOB ?')
                    args.append(f'{f}*')
                else:
                    part.append('folder = ?')
                    args.append(f)
            parts.append(f'({" OR ".join(part)})')

        if self.name:
            parts.append('name GLOB ?')
            args.append(self.name)

        if self.error:
            parts.append('error GLOB ?')
            args.append(self.error)

        return ' AND '.join(f'({part})' for part in parts), args



from __future__ import annotations
from myqueue.workflow import run as run0, wrap, resources


def run(**kwargs):
    return run0(args=['hello', 'world'], **kwargs)


@resources(tmax='1m')
def workflow():
    """Hello world."""
    print(1)
    with run(shell='echo'):
        print(2)
        with run(module='myqueue.test.hello'):
            with resources(tmax='2m'):
                print(3)
                with run(function=print, name='p1'):
                    print(4)
                    with run(script=__file__, tmax='3m'):
                        print(5)
                        with run(script='hello.sh'):
                            print(6)
                            wrap(print)('hello', 'world')


if __name__ == '__main__':
    import sys
    print(*sys.argv[1:])



from __future__ import annotations
import smtplib
import stat
from collections import defaultdict
from email.mime.text import MIMEText
import getpass
from pathlib import Path

from myqueue.config import Configuration
from myqueue.task import Task
from myqueue.utils import mqhome
from myqueue.queue import Queue


def send_notification(queue: Queue,
                      email: str,
                      host: str,
                      username: str = None) -> set[Task]:
    """Compose and send email."""
    notifications = set()
    sql = 'state != "q" AND notifications != "" AND user = ?'
    for task in queue.tasks(sql, [queue.config.user]):
        character = task.state.value
        if character in 'dMTFC' and 'r' in task.notifications:
            notifications.add(task)
        if character in task.notifications:
            task.notifications = task.notifications.replace(character, '')
            notifications.add(task)
        if character in 'dMTFC':
            task.notifications = ''
    if notifications:
        count: dict[str, int] = defaultdict(int)
        lines = []
        for task in notifications:
            name = task.state.name
            count[name] += 1
            lines.append(f'{name}: {task}')
        subject = 'MyQueue: ' + ', '.join(f'{c} {name}'
                                          for name, c in count.items())
        body = '\n'.join(lines)
        password_file = mqhome() / f'.myqueue/{host}'
        password = read_password(password_file)
        send_mail(subject, body,
                  to=email,
                  fro=email,
                  host=host,
                  username=username or email,
                  password=password)
        with queue.connection as con:
            con.executemany('UPDATE tasks SET notifications = ? WHERE id = ?',
                            [(task.notifications, task.id)
                             for task in notifications])
    return notifications


def send_mail(subject: str,
              body: str,
              to: str,
              fro: str,
              host: str,
              username: str,
              password: str) -> None:
    """Send an email.

    >>> send_mail('MyQueue: bla-bla',
    ...          'Hi!\\nHow are you?\\n',
    ...          'you@myqueue.org',
    ...          'me@myqueue.org',
    ...          'test.smtp.org',
    ...          'me',
    ...          '********')
    """
    msg = MIMEText(body)
    msg['Subject'] = subject
    msg['From'] = fro
    msg['To'] = to
    data = msg.as_string()
    if host != 'test.smtp.org':
        with smtplib.SMTP_SSL(host) as s:
            s.login(username, password)
            s.sendmail(msg['From'], [to], data)


def read_password(password_file: Path) -> str:
    """Read password from password file."""
    assert password_file.stat().st_mode & (stat.S_IRGRP | stat.S_IROTH) == 0
    password = password_file.read_text().splitlines()[0]
    return password


def configure_email(config: Configuration) -> None:
    """Check configuration and ask for password if it's not already known."""
    ndct = config.notifications
    if 'email' not in ndct or 'host' not in ndct:
        raise ValueError(
            "Please add 'notifications': "
            "{'email': ..., 'host': ..., 'username': ...} "
            'to your ~/.myqueue/config.py '
            '(username can be left out if it is the same as email).')

    host = ndct['host']
    password_file = mqhome() / f'.myqueue/{host}'
    if password_file.is_file():
        password = read_password(password_file)
    else:
        password = getpass.getpass()
        padding = (len(password) % 80) * '.'
        password_file.write_text(f'{password}\n{padding}\n')
        password_file.chmod(0o600)
        if input('Would you like to send a test email? [yes] ') in {'', 'yes'}:
            send_mail('Test email from myqueue',
                      'Testing ...\n',
                      ndct['email'],
                      ndct['email'],
                      host,
                      ndct.get('username', ndct['email']),
                      password)
            print('Sent!')



from __future__ import annotations


def parse_stderr(text: str) -> tuple[str, bool]:
    r"""Find error message in stderr text and check if it was an OOM error.

    >>> parse_stderr('OOM-kill')
    ('OOM-kill', True)
    >>> parse_stderr('raise SCFConvergenceError\n'
    ...              'Set MCA parameter "orte_base_help_aggregate" '
    ...              'to 0 to see all help / error messages')
    ('raise SCFConvergenceError', False)
    >>> parse_stderr('Bla-bla')
    ('Bla-bla', False)
    >>> parse_stderr('')
    ('-', False)
    """
    lines = text.splitlines()
    for line in lines[::-1]:
        ll = line.lower()
        if any(x in ll for x in ['error:', 'memoryerror', 'malloc',
                                 'memory limit', 'oom-kill',
                                 'out of memory', 'assertionerror']):
            oom = (ll.endswith('memory limit at some point.') or
                   'malloc' in ll or
                   line.startswith('MemoryError') or
                   'oom-kill' in ll or
                   line.endswith('out of memory'))
            return line, oom

    for line in lines:
        ll = line.lower()
        if 'error' in ll:
            return line, False
    if lines:
        return lines[-1], False

    return '-', False



from __future__ import annotations
import os
import subprocess
import warnings
from math import ceil

from myqueue.task import Task
from myqueue.schedulers import Scheduler, SchedulerError


class SLURM(Scheduler):
    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        nodelist = self.config.nodes
        nodes, nodename, nodedct = task.resources.select(nodelist)

        name = task.cmd.short_name
        sbatch = ['sbatch',
                  f'--partition={nodename}',
                  f'--job-name={name}',
                  f'--time={ceil(task.resources.tmax / 60)}',
                  f'--ntasks={task.resources.cores}',
                  f'--nodes={nodes}',
                  f'--chdir={task.folder}',
                  f'--output={name}.%j.out',
                  f'--error={name}.%j.err']

        extra_args = self.config.extra_args + nodedct.get('extra_args', [])
        if extra_args:
            sbatch += extra_args

        features = nodedct.get('features')
        if features:
            warnings.warn('Please use extra_args instead of features!')
            sbatch.append(f'--constraint={features}')

        reservation = nodedct.get('reservation')
        if reservation:
            warnings.warn('Please use extra_args instead of reservation!')
            sbatch.append(f'--reservation={reservation}')

        if task.dtasks:
            ids = ':'.join(str(tsk.id) for tsk in task.dtasks)
            sbatch.append(f'--dependency=afterok:{ids}')

        env = []

        cmd = str(task.cmd)
        if task.resources.processes > 1:
            if 'OMP_NUM_THREADS' not in os.environ:
                env.append(('OMP_NUM_THREADS', '1'))
            mpiexec = self.config.mpiexec
            if 'mpiargs' in nodedct:
                mpiexec += ' ' + nodedct['mpiargs']
            cmd = mpiexec + ' ' + cmd.replace('python3',
                                              self.config.parallel_python)

        # Use bash for the script
        script = '#!/bin/bash -l\n'

        # Add environment variables
        if len(env) > 0:
            script += '\n'.join(f'export {name}={val}' for name, val in env)
            script += '\n'

        home = self.config.home
        script += (
            'id=$SLURM_JOB_ID\n'
            f'mq={home}/.myqueue/slurm-$id\n')

        script += self.get_venv_activation_line()

        script += (
            '(touch $mq-0 && \\\n'
            f' cd {str(task.folder)!r} && \\\n'
            f' {cmd} && \\\n'
            ' touch $mq-1) || \\\n'
            '(touch $mq-2; exit 1)\n')

        if dry_run:
            if verbose:
                print(' \\\n    '.join(sbatch))
                print(script)
            return 1

        # Use a clean set of environment variables without any MPI stuff:
        p = subprocess.run(sbatch,
                           input=script.encode(),
                           capture_output=True,
                           env=os.environ)

        if p.returncode:
            raise SchedulerError((p.stderr + p.stdout).decode())

        return int(p.stdout.split()[-1].decode())

    def cancel(self, id: int) -> None:
        subprocess.run(['scancel', str(id)])

    def hold(self, id: int) -> None:
        subprocess.run(['scontrol', 'hold', str(id)])

    def release_hold(self, id: int) -> None:
        subprocess.run(['scontrol', 'release', str(id)])

    def get_ids(self) -> set[int]:
        user = os.environ.get('USER', 'test')
        cmd = ['squeue', '--user', user]
        p = subprocess.run(cmd, stdout=subprocess.PIPE)
        queued = {int(line.split()[0].decode())
                  for line in p.stdout.splitlines()[1:]}
        return queued

    def maxrss(self, id: int) -> int:
        cmd = ['sacct',
               '-j', str(id),
               '-n',
               '--units=K',
               '-o', 'MaxRSS']
        try:
            p = subprocess.run(cmd, stdout=subprocess.PIPE)
        except FileNotFoundError:
            return 0
        mem = 0
        for line in p.stdout.splitlines():
            line = line.strip()
            if line.endswith(b'K'):
                mem = max(mem, int(line[:-1]) * 1000)
        return mem

    def get_config(self, queue: str = '') -> tuple[list[tuple[str, int, str]],
                                                   list[str]]:
        cmd = ['sinfo',
               '--noheader',
               '-O',
               'CPUs,Memory,Partition']
        p = subprocess.run(cmd, stdout=subprocess.PIPE)
        nodes = []
        for line in p.stdout.decode().splitlines():
            cores, mem, name = line.split()
            nodes.append((name.rstrip('*'),
                          int(cores),
                          mem.rstrip('+') + 'M'))
        return nodes, []



from pathlib import Path
from types import SimpleNamespace

import pytest
from myqueue.states import State
from myqueue.task import create_task


def test_task(tmp_path):
    t = create_task('x', folder=tmp_path)
    assert t.name == 'x.0'
    assert t.id == 0

    t.state = State.running
    assert t.running_time(1.0) == 1.0

    assert repr(t) == 'Task(x)'

    for c in 'ifnAraste':
        x = t.order_key(c)
        assert not (x < x)

    with pytest.raises(ValueError):
        t.order_key('x')

    dct = {'id': 0,
           'folder': str(tmp_path),
           'cmd': {'args': [], 'type': 'python-module', 'cmd': 'x'},
           'state': 'running',
           'resources': {'cores': 1},
           'restart': 0,
           'workflow': False,
           'deps': [],
           'notifications': '',
           'creates': [],
           'tqueued': 0.0,
           'trunning': 0.0,
           'tstop': 0.0,
           'error': '',
           'user': t.user}
    assert t.todict() == dct

    del dct['creates']
    del dct['restart']
    t.fromdict(dct, Path())

    err = tmp_path / 'x.err'

    def oom():
        return t.read_error_and_check_for_oom(
            SimpleNamespace(error_file=lambda _: err))  # type: ignore

    assert not oom()
    err.write_text('... memory limit at some point.')
    assert oom()
    err.write_text('... malloc ...')
    assert oom()
    err.write_text('MemoryError ...')
    assert oom()
    err.write_text('... oom-kill ...')
    assert oom()
    err.write_text('... out of memory')
    assert oom()
    err.write_text('... some other error ...')
    assert not oom()



from __future__ import annotations
import argparse
import os
import sys
import textwrap
from pathlib import Path
from time import time


class MQError(Exception):
    """For nice (expected) CLI errors."""


def error(*args: str) -> None:
    """Write error message to stderr in red."""
    if sys.stderr.isatty():
        print('\033[91m', end='', file=sys.stderr)
        print(*args, '\033[0m', file=sys.stderr)
    else:
        print(*args, file=sys.stderr)


main_description = """\
Frontend for SLURM/LSF/PBS.

Type "mq help <command>" for help.
See https://myqueue.readthedocs.io/ for more information.
"""

HELP = [
    ('help',
     'Show how to use this tool.', """
More help can be found here: https://myqueue.readthedocs.io/.
"""),
    ('list',
     'List tasks in queue.', """
Only tasks in the chosen folder and its subfolders are shown.

Columns:

    i: id
    f: folder
    n: name of task
    a: arguments
    I: info: "+<nargs>,*<repeats>,d<ndeps>"
    r: resources
    A: age
    s: state
    t: time
    e: error message

Examples:

    $ mq list -s rq  # show running and queued jobs
    $ mq ls -s F abc/  # show failed jobs in abc/ folder
"""),
    ('submit',
     'Submit task(s) to queue.', """
Example:

    $ mq submit script.py -R 24:1d  # 24 cores for 1 day
"""),
    ('resubmit',
     'Resubmit done, failed or timed-out tasks.', """
Example:

    $ mq resubmit -i 4321  # resubmit job with id=4321
"""),
    ('remove',
     'Remove or cancel task(s).', """
Examples:

    $ mq remove -i 4321,4322  # remove jobs with ids 4321 and 4322
    $ mq rm -s d . -r  # remove done jobs in this folder and its subfolders
"""),
    ('info',
     'Show detailed information about MyQueue or a task.', """
Example:

    $ mq info  # show information about MyQueue
    $ mq info 12345  # show information about task with id=12345
"""),
    ('workflow',
     'Submit tasks from Python script or several scripts matching pattern.',
     """
The script(s) must define a workflow() function as shown here:

    $ cat flow.py
    from myqueue.workflow import run
    def workflow():
        with run(<task1>):
            run(<task2>)
    $ mq workflow flow.py F1/ F2/  # submit tasks in F1 and F2 folders
"""),
    ('kick',
     'Restart T and M tasks (timed-out and out-of-memory).', """
The queue is kicked automatically every ten minutes - so you don't have
to do it manually.
"""),
    ('modify',
     'Modify task(s).', """
The following state changes are allowed: h->q or q->h.
"""),
    ('init',
     'Initialize new queue.', """
This will create a .myqueue/ folder in your current working directory
and copy ~/.myqueue/config.py into it.
"""),
    ('sync',
     'Make sure SLURM/LSF/PBS and MyQueue are in sync.', """
Remove tasks that SLURM/LSF/PBS doesn't know about.  Also removes a task
if its corresponding folder no longer exists.
"""),
    ('completion',
     'Set up tab-completion for Bash.', """
Do this:

    $ mq completion >> ~/.bashrc
"""),
    ('config',
     'Create config.py file.', """
This tool will try to guess your configuration.  Some hand editing
afterwards will most likely be needed.
Read more about config.py file here:

    https://myqueue.readthedocs.io/en/latest/configuration.html

Example:

    $ mq config -Q hpc lsf
"""),
    ('daemon',
     'Interact with the background process.', """
Manage daemon for sending notifications, restarting, holding and
releasing tasks.
""")]

aliases = {'rm': 'remove',
           'ls': 'list'}


commands: dict[str, tuple[str, str]] = {}
for _cmd, _help, _description in HELP:
    _description = _help + '\n\n' + _description[1:]
    commands[_cmd] = (_help, _description)


def main(arguments: list[str] = None) -> None:
    sys.exit(_main(arguments))


def _main(arguments: list[str] = None) -> int:
    is_test = bool(os.environ.get('MYQUEUE_TESTING'))
    parser = argparse.ArgumentParser(
        prog='mq',
        formatter_class=Formatter,
        description=main_description,
        allow_abbrev=False)
    parser.add_argument('-V', '--version', action='store_true',
                        help='Show version number')

    subparsers = parser.add_subparsers(title='Commands', dest='command')

    short_options: dict[str, int] = {}
    long_options: dict[str, int] = {}

    for cmd, (help, description) in commands.items():
        p = subparsers.add_parser(cmd,
                                  description=description,
                                  help=help,
                                  formatter_class=Formatter,
                                  aliases=[alias for alias in aliases
                                           if aliases[alias] == cmd])

        def a(*args, **kwargs):  # type: ignore
            """Wrapper for Parser.add_argument().

            Hack to fix argparse's handling of options.  See
            fix_option_order() function below."""

            x = p.add_argument(*args, **kwargs)
            if x is None:
                return
            for o in x.option_strings:
                nargs = x.nargs if x.nargs is not None else 1
                assert isinstance(nargs, int)
                if o.startswith('--'):
                    long_options[o] = nargs
                else:
                    short_options[o[1]] = nargs

        if cmd == 'help':
            a('cmd', nargs='?', help='Subcommand.')
            continue

        if cmd == 'daemon':
            a('action', choices=['start', 'stop', 'status'],
              help='Start, stop or check status.')
            a('folder',
              nargs='?',
              help='Pick daemon process corresponding to this folder.  '
              'Defaults to current folder.')

        elif cmd == 'submit':
            a('task', help='Task to submit.')
            a('-d', '--dependencies', default='',
              help='Comma-separated task names.')
            a('-n', '--name', help='Name used for task.')
            a('--restart', type=int, default=0, metavar='N',
              help='Restart N times if task times out or runs out of memory. '
              'Time-limit will be doubled for a timed out task and '
              'number of cores will be increased to the next number of nodes '
              'for a task that runs out of memory.')
            a('folder',
              nargs='*',
              help='Submit tasks in this folder.  '
              'Defaults to current folder.')

        elif cmd == 'config':
            a('scheduler', choices=['local', 'slurm', 'pbs', 'lsf'], nargs='?',
              help='Name of scheduler.  Will be guessed if not supplied.')
            a('-Q', '--queue-name', default='',
              help='Name of queue.  May be needed.')
            a('--in-place', action='store_true',
              help='Overwrite ~/.myqueue/config.py file.')

        if cmd in ['submit', 'workflow']:
            a('--max-tasks', type=int, default=1_000_000_000,
              help='Maximum number of tasks to submit.')

        if cmd == 'resubmit':
            a('--remove', action='store_true',
              help='Remove old tasks.')

        if cmd == 'remove':
            a('-f', '--force', action='store_true',
              help='Remove also workflow tasks.')

        if cmd in ['resubmit', 'submit']:
            a('-R', '--resources',
              help='Examples: "8:1h", 8 cores for 1 hour. '
              'Use "m" for minutes, '
              '"h" for hours and "d" for days. '
              '"16:1:30m": 16 cores, 1 process, half an hour. '
              '"1:xeon40:5m":  1 core on "xeon40" for 5 minutes.')
            a('-w', '--workflow', action='store_true',
              help='Write <task-name>.state file when task has finished.')

        if cmd == 'modify':
            a('-E', '--email', default='u', metavar='STATES',
              help='Send email when state changes to one of the specified '
              'states (one or more of the letters: rdFCTMA).')
            a('-N', '--new-state', default='u',
              help='New state (one of the letters: qhrdFCTM).')

        if cmd == 'workflow':
            a('script', help='Submit tasks from workflow script.')
            a('-f', '--force', action='store_true',
              help='Submit also failed tasks.')
            a('-t', '--targets',
              help='Comma-separated target names.  Without any targets, '
              'all tasks will be submitted.')
            a('-p', '--pattern', action='store_true',
              help='Use submit scripts matching "script" pattern in all '
              'subfolders.')
            a('folder',
              nargs='*',
              help='Submit tasks in this folder.  '
              'Defaults to current folder.')
            a('-a', '--arguments',
              help='Pass arguments to workflow() function.  Example: '
              '"-a name=hello,n=5" will call '
              "workflow(name='hello', n=5).")

        if cmd in ['list', 'remove', 'resubmit', 'modify']:
            a('-s', '--states', metavar='qhrdFCTMaA',
              help='Selection of states. First letters of "queued", "hold", '
              '"running", "done", "FAILED", "CANCELED", "TIMEOUT", '
              '"MEMORY", "all" and "ALL".')
            a('-i', '--id', help="Comma-separated list of task ID's. "
              'Use "-i -" for reading ID\'s from stdin '
              '(one ID per line; extra stuff after the ID will be ignored).')
            a('-n', '--name',
              help='Select only tasks with names matching "NAME" '
              '(* and ? can be used).')
            a('-e', '--error',
              help='Select only tasks with error message matching "ERROR" '
              '(* and ? can be used).')

        if cmd == 'list':
            a('-c', '--columns', metavar='ifnaIrAste', default='ifnaIrAste',
              help='Select columns to show.  Use "-c a-" to remove the '
              '"a" column.')
            a('-S', '--sort', metavar='c', default='i',
              help='Sort rows using column c, where c must be one of '
              'i, f, n, a, r, A, s, t or e.  '
              'Use "-S c-" for a descending sort.')
            a('-C', '--count', action='store_true',
              help='Just show the number of tasks.')
            a('--not-recursive', action='store_true',
              help='Do not list subfolders.')
            a('folder',
              nargs='*',
              help='List tasks in this folder and its subfolders.  '
              'Defaults to current folder.  '
              'Use --not-recursive to exclude subfolders.')

        if cmd not in ['list', 'completion', 'info', 'test']:
            a('-z', '--dry-run',
              action='store_true',
              help='Show what will happen without doing anything.')

        a('-v', '--verbose', action='count', default=0, help='More output.')
        a('-q', '--quiet', action='count', default=0, help='Less output.')
        a('-T', '--traceback', action='store_true',
          help='Show full traceback.')

        if cmd in ['remove', 'resubmit', 'modify']:
            a('-r', '--recursive', action='store_true',
              help='Use also subfolders.')
            a('folder',
              nargs='*',
              help='Task-folder.  Use --recursive (or -r) to include '
              'subfolders.')

        if cmd in ['sync', 'kick']:
            a('folder',
              nargs='?',
              help=f'{cmd.title()} tasks in this folder and its subfolders.  '
              'Defaults to current folder.')

        if cmd == 'info':
            a('-i', '--id',
              help='Show information about specific task.')
            a('-A', '--all', action='store_true',
              help='Show information about all your queues.')
            a('folder',
              nargs='?',
              help='Show information for queues in this folder and its '
              'subfolders.  Defaults to current folder.')

    args = parser.parse_args(
        fix_option_order(arguments if arguments is not None else sys.argv[1:],
                         short_options,
                         long_options))

    args.command = aliases.get(args.command, args.command)

    # Create ~/.myqueue/ if it's not there:
    f = Path.home() / '.myqueue'
    if not f.is_dir():
        f.mkdir()

    if args.version:
        from myqueue import __version__
        print('Version:', __version__)
        print('Code:   ', Path(__file__).parent)
        return 0

    if args.command is None:
        parser.print_help()
        return 0

    if args.command == 'help':
        if args.cmd is None:
            parser.print_help()
        else:
            subparsers.choices[args.cmd].print_help()
        return 0

    if args.command == 'config':
        from myqueue.config import guess_configuration
        guess_configuration(args.scheduler, args.queue_name, args.in_place)
        return 0

    if args.command == 'completion':
        py = sys.executable
        filename = Path(__file__).with_name('complete.py')
        cmd = f'complete -o default -C "{py} {filename}" mq'
        if args.verbose:
            print('Add tab-completion for Bash by copying the following '
                  f'line to your ~/.bashrc (or similar file):\n\n   {cmd}\n')
        else:
            print(cmd)
        return 0

    try:
        run(args, is_test)
    except KeyboardInterrupt:
        pass
    except TimeoutError as x:
        lockfile = x.args[0]
        age = time() - lockfile.stat().st_mtime
        error(f'Locked {age:.0f} seconds ago:', lockfile)
        if age > 60:
            error('Try removing the file and report this to the developers!')
    except MQError as x:
        error(*x.args)
        return 1
    except Exception as x:
        if args.traceback:
            from rich.console import Console
            Console().print_exception()
            raise
        else:
            error(f'{x.__class__.__name__}: {x}',
                  f'\nTo get a full traceback, use: mq {args.command} ... -T')
            return 1
    return 0


def run(args: argparse.Namespace, is_test: bool) -> None:
    from myqueue.config import Configuration, find_home_folder
    from myqueue.resources import Resources
    from myqueue.task import task
    from myqueue.queue import Queue
    from myqueue.selection import Selection
    from myqueue.utils import mqhome
    from myqueue.workflow import workflow, prune
    from myqueue.daemon import start_daemon, perform_daemon_action
    from myqueue.states import State
    from myqueue.pretty import pprint

    verbosity = 1 - args.quiet + args.verbose

    if args.command == 'init':
        root = Path.cwd()
        mq = root / '.myqueue'
        if not mq.is_dir():
            print(f'Creating {mq}')
            mq.mkdir()
        path = mqhome() / '.myqueue'
        cfg1 = path / 'config.py'
        cfg2 = mq / 'config.py'
        if cfg1.is_file() and not cfg2.is_file():
            print(f'Copying {cfg1} to {cfg2}')
            cfg2.write_text(cfg1.read_text())
        return

    folder_names: list[str] = []
    if args.command in ['sync', 'kick', 'daemon', 'info']:
        folder_names = [args.folder or '.']
    else:
        folder_names = args.folder or ['.']

    folders = [Path(folder).expanduser().absolute().resolve()
               for folder in folder_names]

    for folder in folders:
        if not folder.is_dir():
            raise MQError('No such folder:', folder)

    if args.command == 'info' and args.all:
        from myqueue.info import info_all
        info_all(folders[0])
        return

    if args.command in ['remove', 'resubmit', 'modify']:
        if not folders:
            if args.id:
                folders = [Path.cwd()]
            else:
                raise MQError('Missing folder!')

    # Find root folder:
    start = folders[0]
    try:
        config = Configuration.read(start)
    except ValueError:
        raise MQError(
            f'The folder {start} is not inside a MyQueue tree.\n'
            'You can create a tree with "cd <root-of-tree>; mq init".')

    home = config.home

    if verbosity > 1:
        print('Root:', home)

    for folder in folders[1:]:
        try:
            folder.relative_to(home)
        except ValueError:
            raise MQError(f'{folder} not inside {home}')
        home2 = find_home_folder(folder)
        if home2 != home:
            raise MQError(
                f'More than one .myqueue/ folder: {home} and {home2}')

    if args.command == 'daemon':
        perform_daemon_action(config, args.action)
        return

    if not is_test:
        try:
            start_daemon(config)
        except PermissionError:
            pass

    if args.command in ['list', 'remove', 'resubmit', 'modify']:
        default = 'qhrdFCTM' if args.command == 'list' else ''
        states = State.str2states(args.states
                                  if args.states is not None
                                  else default)

        ids: set[int] | None = None
        if args.id:
            if args.states is not None:
                raise MQError("You can't use both -i and -s!")
            if args.folder:
                raise ValueError("You can't use both -i and folder(s)!")

            if args.id == '-':
                ids = {int(line.split()[0]) for line in sys.stdin}
            else:
                ids = {int(id) for id in args.id.split(',')}
        elif args.command != 'list' and args.states is None:
            raise MQError('You must use "-i <id>" OR "-s <state(s)>"!')

        selection = Selection(ids,
                              args.name,
                              states,
                              folders,
                              getattr(args, 'recursive',
                                      not getattr(args, 'not_recursive',
                                                  False)),
                              args.error)

    dry_run = getattr(args, 'dry_run', False)
    need_lock = args.command not in ['list', 'info'] and not dry_run
    with Queue(config, need_lock=need_lock, dry_run=dry_run) as queue:
        if args.command == 'list':
            reverse = args.sort.endswith('-')
            column = args.sort.rstrip('-')
            tasks = queue.select(selection)
            pprint(tasks,
                   verbosity=verbosity,
                   columns=args.columns,
                   short=args.count,
                   sort=column,
                   reverse=reverse)

        elif args.command == 'remove':
            from myqueue.remove import remove
            tasks = queue.select(selection)
            remove(queue, tasks, verbosity, args.force)

        elif args.command == 'resubmit':
            from myqueue.resubmit import resubmit
            resources: Resources | None
            if args.resources:
                resources = Resources.from_string(args.resources)
            else:
                resources = None
            resubmit(queue, selection, resources,
                     remove=args.remove)

        elif args.command == 'submit':
            from myqueue.submitting import submit
            newtasks = [task(args.task,
                             resources=args.resources,
                             name=args.name,
                             folder=str(folder),
                             deps=args.dependencies,
                             workflow=args.workflow,
                             restart=args.restart)
                        for folder in folders]

            submit(queue, newtasks,
                   max_tasks=args.max_tasks,
                   verbosity=verbosity)

        elif args.command == 'modify':
            from myqueue.modify import modify
            state = State(args.new_state)
            modify(queue, selection, state, State.str2states(args.email))

        elif args.command == 'workflow':
            from myqueue.submitting import submit
            tasks = workflow(args, folders, verbosity)
            tasks, done = prune(tasks, queue, args.force)
            try:
                submit(queue, tasks, done=done, max_tasks=args.max_tasks)
            except Exception as ex:
                raise MQError(ex.args)

        elif args.command == 'sync':
            from myqueue.syncronize import sync
            sync(queue)

        elif args.command == 'kick':
            from myqueue.kick import kick
            kick(queue, verbosity)

        else:
            from myqueue.info import info
            assert args.command == 'info'
            info(queue, args.id, verbosity)


class Formatter(argparse.HelpFormatter):
    """Improved help formatter."""
    def _fill_text(self, text: str, width: int, indent: str) -> str:
        assert indent == ''
        out = ''
        blocks = text.split('\n\n')
        for block in blocks:
            if block[0] == '*':
                # List items:
                for item in block[2:].split('\n* '):
                    out += textwrap.fill(item,
                                         width=width - 2,
                                         initial_indent='* ',
                                         subsequent_indent='  ') + '\n'
            elif block[0] == ' ':
                # Indented literal block:
                out += block + '\n'
            else:
                # Block of text:
                out += textwrap.fill(block, width=width) + '\n'
            out += '\n'
        return out[:-1]


def fix_option_order(arguments: list[str],
                     short_options: dict[str, int],
                     long_options: dict[str, int]) -> list[str]:
    """Allow intermixed options and arguments."""
    args1: list[str] = []
    args2: list[str] = []
    i = 0
    while i < len(arguments):
        a = arguments[i]
        if a == '--':
            args2 += arguments[i:]
            break
        if a == '-V' or a == '--version':
            args1.append(a)
        elif a in long_options:
            n = long_options[a]
            args2 += arguments[i:i + 1 + n]
            i += n
        elif a.startswith('--') and '=' in a:
            args2.append(a)
        elif a.startswith('-'):
            for j, b in enumerate(a[1:]):
                n = short_options.get(b, 0)
                if n:
                    if j < len(a) - 2:
                        n = 0
                    args2 += arguments[i:i + 1 + n]
                    i += n
                    break
            else:
                args2.append(a)
        else:
            args1.append(a)
        i += 1
    return args1 + args2



from myqueue.pretty import pprint
from myqueue.task import create_task
import pytest


@pytest.mark.xfail
def test_pretty(capsys):
    pprint([create_task('abc')] * 10, maxlines=4, columns='i-')
    lines = capsys.readouterr().out.splitlines()
    assert len(lines) == 9



from __future__ import annotations

import ast
import runpy
import argparse
from collections import defaultdict
from functools import partial
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, Sequence, Type, Union

from myqueue.caching import json_cached_function, CacheFileNotFoundError
from myqueue.cli import MQError
from myqueue.commands import (Command, PythonModule, PythonScript,
                              ShellCommand, ShellScript, WorkflowTask)
from myqueue.resources import Resources
from myqueue.states import State
from myqueue.task import UNSPECIFIED, Task
from myqueue.utils import chdir, normalize_folder
from myqueue.queue import Queue

DEFAULT_VERBOSITY = 1


def workflow(args: argparse.Namespace,
             folders: list[Path],
             verbosity: int = DEFAULT_VERBOSITY) -> list[Task]:
    """Collect tasks from workflow script(s) and folders."""
    if args.arguments:
        kwargs = str2kwargs(args.arguments)
    else:
        kwargs = {}

    if args.pattern:
        pattern = args.script
        tasks = workflow_from_scripts(pattern,
                                      kwargs,
                                      folders,
                                      verbosity=verbosity)
    else:
        tasks = workflow_from_script(Path(args.script),
                                     kwargs,
                                     folders,
                                     verbosity=verbosity)

    if args.targets:
        names = args.targets.split(',')
        tasks = filter_tasks(tasks, names)

    dnames = set()
    for task in tasks:
        if task.dname in dnames:
            raise MQError('Please use unique names for all tasks.  '
                          f'{task.cmd.name!r} is already used')
        dnames.add(task.dname)

    return tasks


def prune(tasks: Sequence[Task],
          queue: Queue,
          force: bool = False) -> tuple[list[Task], list[Task]]:
    """Only keep tasks that are not already done.

    Will also skip tasks in a bad state (unless *force* is True).
    Done means a task is marked as "done" in the queue or it has created
    its files.
    """
    root = queue.config.home
    ok: list[Task] = []
    done: list[Task] = []
    remove: list[int] = []
    count: defaultdict[str, int] = defaultdict(int)
    for task in tasks:
        # name = str(task.dname.relative_to(root))
        name = task.dname.name
        f = normalize_folder(task.folder, root)
        rows = queue.sql(
            'SELECT id, state FROM tasks WHERE name = ? AND folder = ?',
            [name, f])
        id, state = max(rows, default=(-1, 'u'))
        if id == -1:
            if task.check_creates_files():
                state = 'd*'
                done.append(task)
            else:
                ok.append(task)
        elif force and state in 'FTMC':
            ok.append(task)
            remove.append(id)

        count[state] += 1

    queue.remove(remove)

    if count:
        for state, n in count.items():
            if state == 'u':
                state = 'new'
            elif state == 'd*':
                state = 'done*'
            else:
                state = State(state).name
            print(f'{state:9}: {n}')
    if not force and any(state in 'FTMC' for state in count):
        print('Use --force to submit failed tasks.')

    return ok, done


WorkflowFunction = Callable[[], None]


def get_workflow_function(path: Path,
                          kwargs: dict[str, Any] = {}) -> WorkflowFunction:
    """Get workflow function from script."""
    module = runpy.run_path(str(path))  # type: ignore # bug in typeshed?
    try:
        func = module['workflow']
    except KeyError:
        func = module['create_tasks']
    if kwargs:
        name = func.__name__
        func = partial(func, **kwargs)
        func.__name__ = name
    return func


def workflow_from_scripts(
        pattern: str,
        kwargs: dict[str, Any],
        folders: list[Path],
        verbosity: int = DEFAULT_VERBOSITY) -> list[Task]:
    """Generate tasks from workflows defined by '**/*{script}'."""
    import rich.progress as progress

    tasks: list[Task] = []
    paths = [path
             for folder in folders
             for path in folder.glob('**/*' + pattern)]

    with progress.Progress('[progress.description]{task.description}',
                           progress.BarColumn(),
                           progress.MofNCompleteColumn()) as pb:
        id = pb.add_task('Reading scripts:', total=len(paths))
        for path in paths:
            func = get_workflow_function(path, kwargs)
            tasks += get_tasks_from_folder(path.parent, func, path.absolute())
            pb.advance(id)

    return tasks


def workflow_from_script(script: Path,
                         kwargs: dict[str, Any],
                         folders: list[Path],
                         verbosity: int = DEFAULT_VERBOSITY) -> list[Task]:
    """Collect tasks from workflow defined in python script."""
    import rich.progress as progress

    func = get_workflow_function(script, kwargs)

    tasks: list[Task] = []

    with progress.Progress('[progress.description]{task.description}',
                           progress.BarColumn(),
                           progress.MofNCompleteColumn()) as pb:
        id = pb.add_task('Scanning folders:', total=len(folders))
        for folder in folders:
            tasks += get_tasks_from_folder(folder, func, script.absolute())
            pb.advance(id)

    return tasks


def filter_tasks(tasks: list[Task], names: list[str]) -> list[Task]:
    """Filter tasks that are not in names or in dependencies of names."""
    include = set()
    map = {task.dname: task for task in tasks}
    for task in tasks:
        if task.cmd.name in names:
            for t in task.ideps(map):
                include.add(t)
    filteredtasks = list(include)
    return filteredtasks


def str2kwargs(args: str) -> dict[str, int | str | bool | float]:
    """Convert str to dict.

    >>> str2kwargs('name=hello,n=5')
    {'name': 'hello', 'n': 5}
    """
    kwargs = {}
    for arg in args.split(','):
        key, value = arg.split('=')
        try:
            v = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            v = value
        kwargs[key] = v
    return kwargs


def get_tasks_from_folder(folder: Path,
                          func: Callable,
                          script: Path) -> list[Task]:
    """Collect tasks from folder."""
    with chdir(folder):
        if func.__name__ == 'create_tasks':
            tasks = func()
            for task in tasks:
                task.workflow = True
        else:
            tasks = collect(func, script)

    return tasks


class StopCollecting(Exception):
    """Workflow needs an actual result instead of a dummy Result object."""


class StopRunning(Exception):
    """The correct task has finished running: Stop workflow function."""


class RunHandle:
    """Result of calling run().  Can be used as a context manager."""
    def __init__(self, task: Task, runner: Runner):
        self.task = task
        self.runner = runner

    @property
    def result(self) -> Result:
        """Result from Python-function tasks."""
        result = self.task.result
        if result is UNSPECIFIED:
            return Result(self.task)
        assert not isinstance(result, str)
        return result

    @property
    def done(self) -> bool:
        """Has task been successfully finished?"""
        return self.task.state == State.done

    def __enter__(self) -> RunHandle:
        self.runner.dependencies.append(self.task)
        return self

    def __exit__(self,
                 exc_type: Type[BaseException],
                 exc_val: BaseException,
                 exc_tb: TracebackType) -> None:
        task = self.runner.dependencies.pop()
        assert task is self.task


class Result:
    """Result object for a task - used for finding dependencies."""
    def __init__(self, task: Task):
        self.task = task

    def __getattr__(self, attr: str) -> Result:
        return self

    def __lt__(self, other: Any) -> bool:
        raise StopCollecting

    __gt__ = __lt__

    __getitem__ = __getattr__
    __add__ = __getattr__


def get_name(func: Callable) -> str:
    """Give a name to a function.

    >>> import time
    >>> get_name(time.sleep)
    'time.sleep'
    """
    mod = func.__module__
    if mod in ['<run_path>', '__main__']:
        return func.__name__
    return f'{mod}.{func.__name__}'


class ResourceHandler:
    """Resource decorator and context manager."""
    def __init__(self, kwargs: dict[str, Any], runner: Runner):
        self.kwargs = kwargs
        self.runner = runner
        self.old_kwargs: dict

    def __call__(self, workflow_function: Callable[[], Any]
                 ) -> Callable[[], Any]:
        def new() -> Any:
            with self:
                return workflow_function()
        return new

    def __enter__(self) -> None:
        self.old_kwargs = self.runner.resource_kwargs
        self.runner.resource_kwargs = {**self.old_kwargs, **self.kwargs}

    def __exit__(self,
                 exc_type: Type[BaseException],
                 exc_val: BaseException,
                 exc_tb: TracebackType) -> None:
        self.runner.resource_kwargs = self.old_kwargs


class Runner:
    """Wrapper for collecting tasks from workflow function."""
    def __init__(self) -> None:
        self.tasks: list[Task] | None = None
        self.dependencies: list[Task] = []
        self.resource_kwargs = {'tmax': '10m',
                                'cores': 1,
                                'nodename': '',
                                'processes': 0,
                                'restart': 0,
                                'weight': -1}
        self.target = ''
        self.workflow_script: Path | None = None

    def run(self,
            *,
            function: Callable = None,
            script: Union[Path, str] = None,
            module: str = None,
            shell: str = None,
            name: str = '',
            args: Sequence[Any] = [],
            kwargs: dict[str, Any] = {},
            deps: list[RunHandle] = [],
            creates: list[str] = [],
            tmax: str = None,
            cores: int = None,
            nodename: str = None,
            processes: int = None,
            restart: int = None,
            weight: float = None,
            folder: Path | str = '.') -> RunHandle:
        """Run or submit a task.

        The type of task is specifyed by one and only one of the arguments
        `function`, `script`, `module` or `shell`.

        Parameters
        ----------
        function:
            A Python function.
        script:
            Shell-script or Python-script.
        module:
            Name of a Python module.
        shell:
            Shell command.  Must be found in $PATH.
        args: list of objects
            Command-line arguments or function arguments.
        name: str
            Name to use for task.  Default is name of function, script, module
            or shell command.
        deps: list of run handle objects
            Dependencies.
        creates: list of filenames
            Files created by task.
        cores: int
            Number of cores (default is 1).
        nodename: str
            Name of node.
        processes: int
            Number of processes to start (default is one for each core).
        tmax: str
            Maximum time for task.  Examples: "40s", "30m", "20h" and "2d".
        folder: str
            Folder where task should run (default is current folder).
        restart: int
            How many times to restart task.
        weight: float
            Weight of task.  See :ref:`task_weight`.

        Returns
        -------
        RunHandle
            Object representing the run.
        """
        dependencies = self.extract_dependencies(args, kwargs, deps)

        resource_kwargs = {}
        values = [tmax, cores, nodename, processes, restart, weight]
        for value, (key, default) in zip(values,
                                         self.resource_kwargs.items()):
            resource_kwargs[key] = value if value is not None else default

        task = create_task(function,
                           script,
                           module,
                           shell,
                           name,
                           args,
                           kwargs,
                           dependencies,
                           self.workflow_script,
                           Path(folder).absolute(),
                           resource_kwargs.pop('restart'),  # type: ignore
                           creates=creates,
                           **resource_kwargs)

        if self.target:
            if task.cmd.fname == self.target:
                task.run()
                raise StopRunning
        elif self.tasks is not None:
            if task.state != State.done:
                self.tasks.append(task)
        else:
            task.run()

        return RunHandle(task, self)

    def extract_dependencies(self,
                             args: Sequence[Any],
                             kwargs: dict[str, Any],
                             deps: list[RunHandle]) -> list[Path]:
        """Find dependencies on other tasks."""
        tasks = set(self.dependencies)
        for handle in deps:
            tasks.add(handle.task)
        for thing in list(args) + list(kwargs.values()):
            if isinstance(thing, Result):
                tasks.add(thing.task)
        return [task.dname for task in tasks]

    def wrap(self, function: Callable, **run_kwargs: Any) -> Callable:
        """Wrap a function as a task.

        Takes the same keyword arguments as `run`
        (except module, script and shell).

        These two are equivalent::

            result = run(function=func, args=args, kwargs=kwargs, ...).result
            result = wrap(func, ...)(*args, **kwargs)

        """
        def wrapper(*args: Any, **kwargs: Any) -> Result:
            handle = self.run(function=function,
                              args=args,
                              kwargs=kwargs,
                              **run_kwargs)
            return handle.result
        return wrapper

    def resources(self,
                  *,
                  tmax: str = None,
                  cores: int = None,
                  nodename: str = None,
                  processes: int = None,
                  restart: int = None,
                  weight: float = None) -> ResourceHandler:
        """Resource decorator and context manager.

        Parameters
        ----------
        cores: int
            Number of cores (default is 1).
        nodename: str
            Name of node.
        processes: int
            Number of processes to start (default is one for each core).
        tmax: str
            Maximum time for task.  Examples: "40s", "30m", "20h" and "2d".
        restart: int
            How many times to restart task.
        """
        keys = ['tmax', 'cores', 'nodename', 'processes', 'restart', 'weight']
        values = [tmax, cores, nodename, processes, restart, weight]
        kwargs = {key: value
                  for key, value in zip(keys, values)
                  if value is not None}
        return ResourceHandler(kwargs, self)


runner = Runner()
run = runner.run
wrap = runner.wrap
resources = runner.resources


def create_task(function: Callable = None,
                script: Union[Path | str] = None,
                module: str = None,
                shell: str = None,
                name: str = '',
                args: Sequence[Any] = [],
                kwargs: dict[str, Any] = {},
                deps: list[Path] = [],
                workflow_script: Path = None,
                folder: Path = Path('.'),
                restart: int = 0,
                creates: list[str] = [],
                **resource_kwargs: Any) -> Task:
    """Create a Task object."""
    if sum(arg is not None
           for arg in [function, module, script, shell]) != 1:
        1 / 0

    command: Command

    if function:
        name = name or get_name(function)
        cached_function = json_cached_function(function, name, args, kwargs)
        command = WorkflowTask(f'{workflow_script}:{name}', [],
                               cached_function)
        creates = creates + [f'{name}.result']
    elif module:
        assert not kwargs
        command = PythonModule(module, [str(arg) for arg in args])
    elif script:
        assert not kwargs
        path = folder / script
        assert path.is_file(), path
        if path.suffix == '.py':
            command = PythonScript(str(path), [str(arg) for arg in args])
        else:
            command = ShellScript(str(path), [str(arg) for arg in args])
    else:
        assert not kwargs
        assert isinstance(shell, str)
        command = ShellCommand('shell:' + shell, [str(arg) for arg in args])

    if name:
        command.set_non_standard_name(name)

    res = Resources.from_args_and_command(command=command,
                                          path=folder,
                                          **resource_kwargs)

    task = Task(command,
                deps=deps,
                resources=res,
                folder=folder,
                restart=restart,
                workflow=True,
                creates=creates)

    if function and not any(isinstance(thing, Result)
                            for thing in list(args) + list(kwargs.values())):
        try:
            task.result = cached_function(only_read_from_cache=True)
            task.state = State.done
        except CacheFileNotFoundError:
            pass

    return task


def collect(workflow_function: Callable,
            script: Path) -> list[Task]:
    """Collecting tasks from workflow function."""
    runner.tasks = []
    runner.workflow_script = script
    try:
        workflow_function()
    except StopCollecting:
        pass

    tasks = runner.tasks
    runner.tasks = None
    return tasks


def run_workflow_function(script: str, name: str) -> None:
    """Run specific task in workflow function."""
    workflow_function = get_workflow_function(Path(script))
    runner.target = name
    try:
        workflow_function()
    except StopRunning:
        pass
    runner.target = ''



from __future__ import annotations
from myqueue.resources import Resources
from myqueue.selection import Selection
from myqueue.task import Task
from myqueue.queue import Queue
from myqueue.submitting import submit


def resubmit(queue: Queue,
             selection: Selection,
             resources: Resources | None,
             remove: bool = False) -> None:
    """Resubmit tasks."""
    tasks = []
    ids = []
    for task in queue.select(selection):
        if task.state in {'queued', 'hold', 'running'}:
            continue

        ids.append(task.id)
        task = Task(task.cmd,
                    deps=task.deps,
                    resources=resources or task.resources,
                    folder=task.folder,
                    restart=task.restart,
                    workflow=task.workflow,
                    creates=task.creates)
        tasks.append(task)

    if remove and not queue.dry_run:
        queue.remove(ids)

    submit(queue, tasks)



from __future__ import annotations
import asyncio
import pickle
import socket
from functools import partial
from typing import Any

from myqueue.schedulers import Scheduler
from myqueue.task import Task
from myqueue.config import Configuration
from myqueue.states import State
from myqueue.queue import Queue


class LocalSchedulerError(Exception):
    pass


class LocalScheduler(Scheduler):
    port = 39999

    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        if dry_run:
            if verbose:
                print(task)
            return 1
        task.cmd.function = None
        id = self.send('submit', task)
        return id

    def cancel(self, id: int) -> None:
        self.send('cancel', id)

    def hold(self, id: int) -> None:
        self.send('hold', id)

    def release_hold(self, id: int) -> None:
        self.send('release', id)

    def get_ids(self) -> set[int]:
        ids = self.send('list')
        return ids

    def send(self, *args: Any) -> Any:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect(('127.0.0.1', self.port))
            except ConnectionRefusedError:
                raise ConnectionRefusedError(
                    'Local scheduler not responding.  '
                    'Please start it with:\n\n'
                    '    python3 -m myqueue.schedulers.local')
            b = pickle.dumps(args)
            assert len(b) < 4096
            s.sendall(b)
            if args[0] == 'stop':
                return
            b = b''.join(iter(partial(s.recv, 4096), b''))
        status, result = pickle.loads(b)
        if status != 'ok':
            raise LocalSchedulerError(status)
        return result

    def get_config(self, queue: str = '') -> tuple[list[tuple[str, int, str]],
                                                   list[str]]:
        return [], []


class Server:
    """Run tasks in subprocesses."""
    def __init__(self,
                 config: Configuration,
                 cores: int = 1,
                 port: int = 39999) -> None:
        self.config = config
        self.port = port

        with Queue(config) as queue:
            maxid = queue.connection.execute(
                'SELECT MAX(id) FROM tasks').fetchone()[0] or 0
            self.next_id = 1 + maxid

        self.tasks: dict[int, Task] = {}
        self.processes: dict[int, asyncio.subprocess.Process] = {}
        self.aiotasks: dict[int, asyncio.Task] = {}
        self.folder = self.config.home / '.myqueue'

    def start(self) -> None:
        """Start server and wait for cammands."""
        try:
            asyncio.run(self._start())
        except asyncio.exceptions.CancelledError:
            pass

    async def _start(self) -> None:
        for _ in range(5):
            try:
                self.server = await asyncio.start_server(
                    self.recv, '127.0.0.1', self.port)
            except OSError:
                self.port -= 1
            else:
                break
        else:  # no break
            raise OSError('Could not find unused port!')

        print('Port:', self.port)

        async with self.server:  # type: ignore
            await self.server.serve_forever()
            await self.server.wait_closed()

    async def recv(self, reader: Any, writer: Any) -> None:
        """Recieve command from client."""
        data = await reader.read(4096)
        cmd, *args = pickle.loads(data)
        print('COMMAND:', cmd, *args, [(task.id,
                                        [d.id for d in task.dtasks])
                                       for task in self.tasks.values()])
        result: Any
        if cmd == 'stop':
            for aiotask in self.aiotasks.values():
                await aiotask
            self.server.close()
            print('BYE')
            return
        elif cmd == 'submit':
            task = args[0]
            task.id = self.next_id
            task.state = State.queued
            print([d.id for d in task.dtasks])
            if all(t.id in self.tasks for t in task.dtasks):
                task.deps = [t.dname for t in task.dtasks]
                self.tasks[task.id] = task
            self.next_id += 1
            result = task.id
        elif cmd == 'list':
            result = list(self.tasks)
        elif cmd == 'cancel':
            id = args[0]
            if id in self.processes:
                self.terminate(id)
            elif id in self.tasks:
                task = self.tasks[id]
                task.state = State.CANCELED
                self.cancel_dependents(task)
            result = None
        else:
            1 / 0
        writer.write(pickle.dumps(('ok', result)))
        await writer.drain()
        writer.close()
        self.kick()
        print('JOBS:', list(self.tasks))

    def kick(self) -> None:
        """Check if a new task should be started."""
        self.aiotasks = {id: aiotask
                         for id, aiotask in self.aiotasks.items()
                         if not aiotask.done()}

        for task in self.tasks.values():
            if task.state == State.running:
                return

        for task in self.tasks.values():
            if task.state == State.queued and not task.deps:
                break
        else:  # no break
            return

        print('START', task.id)
        aiotask = asyncio.create_task(self.run(task))
        self.aiotasks[task.id] = aiotask
        # aiotask.add_done_callback(lambda t: self.aiotasks.pop(task.id))

    async def run(self, task: Task) -> None:
        """Run a task."""
        out = f'{task.cmd.short_name}.{task.id}.out'
        err = f'{task.cmd.short_name}.{task.id}.err'

        cmd = str(task.cmd)
        if task.resources.processes > 1:
            mpiexec = f'{self.config.mpiexec} -x OMP_NUM_THREADS=1 '
            mpiexec += f'-np {task.resources.processes} '
            cmd = mpiexec + cmd.replace('python3',
                                        self.config.parallel_python)
        cmd = f'{cmd} 2> {err} > {out}'

        proc = await asyncio.create_subprocess_shell(
            cmd, cwd=task.folder)

        self.processes[task.id] = proc

        loop = asyncio.get_event_loop()
        tmax = task.resources.tmax
        handle = loop.call_later(tmax, self.terminate, task.id)
        task.state = State.running
        (self.folder / f'local-{task.id}-0').write_text('')  # running

        await proc.wait()
        print('END', task.id, proc.returncode)
        handle.cancel()

        del self.tasks[task.id]
        del self.processes[task.id]

        if proc.returncode == 0:
            for t in self.tasks.values():
                if task.dname in t.deps:
                    t.deps.remove(task.dname)
            state = 1
        else:
            if task.state == 'TIMEOUT':
                state = 3
            else:
                state = 2
            self.cancel_dependents(task)

        (self.folder / f'local-{task.id}-{state}').write_text('')

        self.kick()

    def _cancel_dependents(self, task: Task) -> None:
        for job in self.tasks.values():
            print('CANCEL', len(self.tasks), task.dname, job, job.deps)
            if task.dname in job.deps:
                job.state = State.CANCELED
                self._cancel_dependents(job)

    def cancel_dependents(self, task: Task) -> None:
        self._cancel_dependents(task)
        self.tasks = {id: task for id, task in self.tasks.items()
                      if task.state != State.CANCELED}

    def terminate(self, id: int, state: State = State.TIMEOUT) -> None:
        """Terminate a task."""
        print('Terminate', id)
        proc = self.processes.get(id)
        if proc and proc.returncode is None:
            proc.terminate()
            self.tasks[id].state = state


if __name__ == '__main__':
    asyncio.run(Server(Configuration.read())._start())



from pathlib import Path
from myqueue.workflow import run


def test_name(mq):
    Path('dir').mkdir()
    mq('submit hello.sh -n helloworld dir')
    mq('ls -n helloworld')
    mq('rm -n helloworld dir -s aA')
    assert mq.wait() == ''


def workflow():
    with run(function=print, name='T1'):
        run(function=print, name='T2')


def test_name_dep(mq):
    dir = Path('dir')
    dir.mkdir()
    wf = Path(__file__)
    mq(f'workflow {wf} dir')
    assert mq.wait() == 'dd'
    mq(f'workflow {wf} dir')
    assert mq.states() == 'dd'
    mq('rm -n T2 -s d dir --force')
    mq('ls')
    mq(f'workflow {wf} dir')
    assert mq.states() == 'd'
    (dir / 'T2.result').unlink()
    mq(f'workflow {wf} dir')
    assert mq.wait() == 'dd'



from __future__ import annotations
from myqueue.resources import Resources


def test_res():
    r = '16:1:xeon8:2h'
    r1 = Resources.from_args_and_command(resources=r)
    r2 = Resources.from_string(r)
    assert str(r1) == str(r2)



from __future__ import annotations
import sys
import subprocess
from pathlib import Path

import pytest

from myqueue.queue import Queue
from myqueue.task import task
from myqueue.submitting import submit


@pytest.mark.skipif(sys.version_info < (3, 9),
                    reason='requires Python 3.9 or higher')
def test_completion():
    from myqueue.utils import update_readme_and_completion
    update_readme_and_completion(test=True)


def test_api(mq):
    from myqueue import submit as simple_submit
    from myqueue.task import task
    simple_submit(task('myqueue.test@oom 1'))
    simple_submit(task('myqueue.test@timeout_once', tmax='1s'))
    mq.wait()
    simple_submit(task('myqueue.test@timeout_once'))
    assert mq.wait() == 'MTd'


def test_backends(mq):
    from myqueue.config import guess_configuration, guess_scheduler
    config = mq.scheduler.config
    config.nodes = [('abc16', {'cores': 16, 'memory': '16G'}),
                    ('abc8', {'cores': 8, 'memory': '8G'})]
    config.mpiexec = 'echo'
    for name in ['slurm', 'lsf', 'pbs']:
        print(name)
        if name == 'pbs':
            p = Path('venv/bin/')
            p.mkdir(parents=True)
            (p / 'activate').write_text('')
        config.scheduler = name
        with Queue(config, dry_run=True) as q:
            submit(q, [task('shell:echo hello', cores=24)])

    try:
        guess_scheduler()
    except ValueError:
        pass  # can happen if more than one out of sbatch, bsub and qsub exist
    guess_configuration('local')


class Result:
    def __init__(self, stdout):
        self.stdout = stdout


def run(commands, stdout):
    if commands[0] == 'sinfo':
        return Result(b'8 256000+ xeon8*\n')
    return Result(b'id state 8:8 load xeon8 128 G\n')


def test_autoconfig(monkeypatch):
    from myqueue.schedulers import get_scheduler
    from myqueue.config import Configuration

    monkeypatch.setattr(subprocess, 'run', run)
    nodes, _ = get_scheduler(Configuration('slurm')).get_config()
    assert nodes == [('xeon8', 8, '256000M')]

    nodes, _ = get_scheduler(Configuration('LSF')).get_config()
    assert nodes == [('xeon8', 8, '128G')]


def test_commands():
    from myqueue.commands import ShellScript, convert, create_command
    assert convert('True') is True
    assert convert('False') is False
    assert convert('3.14') == 3.14
    assert convert('42') == 42
    cmd = create_command('./script.sh 1 2')
    assert isinstance(cmd, ShellScript)
    assert cmd.todict()['args'] == ['1', '2']
    print(cmd)


def test_resource_comments(tmp_path):
    from myqueue.task import task
    script = tmp_path / 'script.py'
    script.write_text('# Script\n# MQ: resources=2:1h\n')
    t = task(str(script))
    assert t.resources.cores == 2
    assert t.resources.tmax == 3600



from __future__ import annotations
import json
from pathlib import Path


def factor(x: int) -> list[int]:
    for f in range(2, x // 2 + 1):
        if x % f == 0:
            return [f] + factor(x // f)
    return [x]


if __name__ == '__main__':
    x = int(Path.cwd().name)  # name of current folder
    factors = factor(x)
    Path('factors.json').write_text(json.dumps({'factors': factors}))



"""Background daemon process.

The daemon process wakes up every ten minutes to check if any tasks need to
be resubmitted, held or released.  Notification emails will also be sent.
It will write its output to .myqueue/daemon.out.
"""

from __future__ import annotations
import functools
import os
import signal
import socket
import sys
import traceback
from pathlib import Path
from time import sleep, time
from typing import Any

from myqueue.queue import Queue
from myqueue.config import Configuration
from myqueue.kick import kick

T = 600  # kick system every ten minutes


def is_running(pidfile: Path) -> bool:
    """Check if daemon is running."""
    if pidfile.is_file():
        age = time() - pidfile.stat().st_mtime
        if age < 7200:
            # No action for two hours - it must be dead:
            return True
    return False


def start_daemon(config: Configuration) -> bool:
    """Fork a daemon process."""
    pidfile = config.home / f'.myqueue/daemon-{config.user}.pid'

    if is_running(pidfile):
        return False

    pid = os.fork()
    if pid == 0:
        if os.getenv('MYQUEUE_TESTING'):
            # Simple version for pytest only:
            loop(config)
        else:
            pid = os.fork()
            if pid == 0:
                # redirect standard file descriptors
                sys.stderr.flush()
                si = open(os.devnull, 'r')
                so = open(os.devnull, 'w')
                se = open(os.devnull, 'w')
                os.dup2(si.fileno(), sys.stdin.fileno())
                os.dup2(so.fileno(), sys.stdout.fileno())
                os.dup2(se.fileno(), sys.stderr.fileno())
                loop(config)
        os._exit(0)
    return True


def exit(pidfile: Path, signum: int, frame: Any) -> None:
    """Remove .myqueue/daemon.pid file on exit."""
    if pidfile.is_file():
        pidfile.unlink()
    if not os.getenv('MYQUEUE_TESTING'):
        sys.exit()


def read_hostname_and_pid(pidfile: Path) -> tuple[str, int]:
    """Read from .myqueue/daemon.pid file."""
    host, pid = pidfile.read_text().split(':')
    return host, int(pid)


def loop(config: Configuration) -> None:
    """Main loop: kick system every ten minutes."""
    err = config.home / f'.myqueue/daemon-{config.user}.err'
    out = config.home / f'.myqueue/daemon-{config.user}.out'
    pidfile = config.home / f'.myqueue/daemon-{config.user}.pid'

    pid = os.getpid()
    host = socket.gethostname()
    pidfile.write_text(f'{host}:{pid}\n')

    cleanup = functools.partial(exit, pidfile)
    signal.signal(signal.SIGWINCH, cleanup)
    signal.signal(signal.SIGTERM, cleanup)

    while True:
        sleep(T)

        if not (config.home / '.myqueue').is_dir():
            break

        try:
            with Queue(config) as queue:
                result = kick(queue, verbosity=0)
        except Exception:
            err.write_text(traceback.format_exc())
            break

        if result:
            with out.open('a') as fd:
                print(result, file=fd)

        pidfile.touch()

    pidfile.unlink()


def perform_daemon_action(config: Configuration, action: str) -> int:
    """Status of, stop or start daemon.

    Returns PID.
    """
    pidfile = config.home / f'.myqueue/daemon-{config.user}.pid'
    pid = -1

    running = is_running(pidfile)
    if running:
        host, pid = read_hostname_and_pid(pidfile)

    if action == 'status':
        if running:
            print(f'Running on {host} with pid={pid}')
        else:
            print('Not running')

    elif action == 'stop':
        if running:
            if host == socket.gethostname():
                try:
                    os.kill(pid, signal.SIGWINCH)
                except ProcessLookupError:
                    pass
            else:
                print(f'You have to be on {host} in order to stop the daemon')
                return 1
        else:
            print('Not running!')

    elif action == 'start':
        if running:
            print('Already running')
        else:
            if pidfile.is_file():
                pidfile.unlink()
            start_daemon(config)
            while not pidfile.is_file():
                # Wait for the fork to start ...
                sleep(0.05)
            host, pid = read_hostname_and_pid(pidfile)
            print(f'PID: {pid}')

    else:
        assert False, action

    return pid



from __future__ import annotations

from myqueue.pretty import pprint
from myqueue.task import Task
from myqueue.utils import plural
from myqueue.queue import Queue


def remove(queue: Queue,
           tasks: list[Task],
           verbosity: int = 1,
           force: bool = False) -> None:
    """Remove or cancel tasks."""
    ntasks0 = len(tasks)
    tasks = [task for task in tasks if force or not task.workflow]
    ntasks = len(tasks)
    if ntasks < ntasks0:
        print(plural(ntasks0 - ntasks, 'task'), 'part of workflow.  '
              'Use --force to remove.')

    if queue.dry_run:
        if tasks:
            pprint(tasks, verbosity=0, sort='i')
            print(plural(len(tasks), 'task'), 'to be removed')
    else:
        for task in tasks:
            if task.state in ['running', 'hold', 'queued']:
                queue.scheduler.cancel(task.id)
        queue.remove((task.id for task in tasks))
        if verbosity > 0:
            if tasks:
                pprint(tasks, verbosity=0, sort='i')
                print(plural(len(tasks), 'task'), 'removed')



from __future__ import annotations
import functools
# import socket
# import time

from myqueue.daemon import perform_daemon_action


def test_daemon(mq, capsys):
    cmd = functools.partial(perform_daemon_action,
                            mq.config)
    cmd('status')
    cmd('stop')
    expected = ['Not running',
                'Not running!']
    # We have to skip this because it forks the test suite which will
    # then run twice!
    """
    host = socket.gethostname()
    pid = cmd('start')
    cmd('start')
    cmd('status')
    cmd('stop')
    time.sleep(0.2)
    cmd('status')
    expected += [
        f'PID: {pid}',
        'Already running',
        f'Running on {host} with pid={pid}',
        'Not running']
    """
    assert capsys.readouterr().out == '\n'.join(expected + [''])



from __future__ import annotations
import pytest
from myqueue.utils import Lock


def test_lock(tmp_path):
    lockfile = tmp_path / 'lock'
    with Lock(lockfile):
        with pytest.raises(TimeoutError):
            with Lock(lockfile, timeout=0.07):
                pass
    assert not lockfile.is_file()



from __future__ import annotations

import os
from pathlib import Path
import time

import pytest
from myqueue.test.flow1 import workflow
from myqueue.test.hello import workflow as hello


def test_flow1(mq):
    script = Path(__file__).with_name('flow1.py')

    mq(f'workflow {script}')
    assert mq.wait() == 'd' * 7

    mq(f'workflow {script}')
    assert mq.wait() == 'd' * 8

    mq('rm -sd')
    assert mq.wait() == 'd' * 8

    mq('rm -sd --force')
    assert mq.wait() == ''

    mq(f'workflow {script}')
    assert mq.wait() == ''

    Path('builtins.print.result').unlink()
    mq(f'workflow {script}')
    assert mq.wait() == 'd'


def test_direct_cached_flow1(tmp_path, capsys):
    os.chdir(tmp_path)
    a = workflow()
    assert a == 3
    assert capsys.readouterr().out == '\n'.join('0213243') + '\n'
    workflow()
    assert capsys.readouterr().out == ''


def test_workflow_old(mq):
    script = Path(__file__)
    mq(f'workflow {script}')


def test_hello(mq):
    Path('hello.sh').write_text('echo $@')
    script = Path(__file__).with_name('hello.py')
    mq(f'workflow {script}')
    assert mq.wait() == 'dddddd'


def test_direct_hello(tmp_path):
    os.chdir(tmp_path)
    Path('hello.sh').write_text('echo $@')
    hello()


def test_flow2(mq):
    script = Path(__file__).with_name('flow2.py')
    mq(f'workflow {script}')
    assert mq.wait() == 'MCCC'
    mq('rm -sC . --force')
    mq(f'workflow {script}', error=1)
    assert mq.states() == 'M'
    mq(f'workflow {script} --force')
    assert mq.wait() == 'MCCC'


wf = """
from myqueue.task import task
def create_tasks():
    t1 = task('shell:sleep+3')
    t2 = task('shell:touch+hello', deps=[t1], creates=['hello'])
    return [t1, t2]
"""


def test_workflow(mq):
    mq('submit shell:sleep+3 -R1:1m -w')
    time.sleep(2)
    Path('wf.py').write_text(wf)
    mq('workflow wf.py . -t shell:touch+hello')
    assert mq.wait() == 'dd'
    mq('workflow wf.py .')
    assert mq.states() == 'dd'
    mq('rm -i 2 -z')  # dry-run
    mq('rm -i 2')  # needs --force
    assert mq.states() == 'dd'
    mq('rm -i 2 --force')
    mq('workflow wf.py .')
    assert mq.wait() == 'd'
    hello = Path('hello')
    hello.unlink()
    mq('workflow wf.py .')
    assert mq.wait() == 'dd'
    assert hello.is_file()


def test_workflow_running_only_with_targets(mq):
    Path('wf.py').write_text(wf)
    mq('workflow wf.py . -t shell:touch+hello')
    assert mq.wait() == 'dd'


def test_workflow_with_failed_job(mq):
    # Introduce bug in workflow task:
    Path('wf.py').write_text(wf.replace('+3', '+a'))
    mq('workflow wf.py .')
    assert mq.wait() == 'FC'

    # Fix bug:
    Path('wf.py').write_text(wf)
    mq('workflow wf.py . --force --dry-run')
    assert mq.wait() == 'FC'

    mq('workflow wf.py . --force')
    assert mq.wait() == 'Fdd'


wf2 = """
from myqueue.task import task
def create_tasks(name, n):
    assert name == 'hello'
    assert n == 5
    return [task('shell:echo+hi', name=f'x{i}', weight=1) for i in range(4)]
"""


def test_workflow2(mq):
    Path('wf2.py').write_text(wf2)
    mq('workflow wf2.py . -a name=hello,n=5')
    mq('kick')
    assert mq.states() == 'hhqq'
    mq.wait()
    mq('kick')
    assert mq.wait() == 'dddd'


wf2_new = """
from myqueue.workflow import run
def workflow():
    for i in range(4):
        run(shell='echo', args=['hi'], name=f'x{i}', weight=1)
"""


def test_workflow2_new(mq):
    Path('wf2.py').write_text(wf2_new)
    mq('workflow wf2.py')
    mq('kick')
    assert mq.states() == 'hhqq'
    mq.wait()
    mq('kick')
    assert mq.wait() == 'dddd'


def test_failing_scheduler(mq):
    with pytest.raises(RuntimeError):
        # Special argument that makes test-scheduler raise an error:
        mq('submit "time.sleep FAIL"')
    assert mq.wait() == ''


wf3 = """
from myqueue.task import task
def create_tasks():
    return [task('shell:echo+hi'),
            task('shell:echo+FAIL')]
"""


def test_workflow3(mq):
    Path('wf3.py').write_text(wf3)
    mq('workflow wf3.py', error=1)
    assert mq.wait() == 'd'


wf4 = """
from myqueue.workflow import run
def workflow():
    with run(function=lambda: None, name='A'):
        run(function=lambda: None, name='B')
"""


def test_workflow_depth_first(mq):
    """Order should be 1/A, 1/B, 2/A, 2/B and not 1/A, 2/A, 1/B, 2/B."""
    Path('wf4.py').write_text(wf4)
    Path('1').mkdir()
    Path('2').mkdir()
    mq('workflow wf4.py 1 2')
    assert mq.wait() == 'dddd'
    assert Path('1/B.2.out').is_file()


wf5 = """
from myqueue.workflow import run
def workflow():
    run(function=lambda: None, name='A')
    run(function=lambda: None, name='A')
"""


def test_workflow_repeated_name(mq):
    Path('wf5.py').write_text(wf5)
    mq('workflow wf5.py', error=1)


wf_creates = """
from myqueue.workflow import run
def workflow():
    with run(shell='echo', creates=['out.txt']):
        run(function=print)
"""


def test_creates(mq):
    Path('wf.py').write_text(wf_creates)
    Path('out.txt').write_text('OK\n')
    mq('workflow wf.py')
    assert mq.wait() == 'd'



from __future__ import annotations

import os
from pathlib import Path

from myqueue.config import Configuration
from myqueue.task import Task


class Scheduler:
    def __init__(self, config: Configuration):
        self.config = config
        self.name = config.scheduler.lower()
        venv = os.environ.get('VIRTUAL_ENV')
        self.activation_script = (Path(venv) / 'bin/activate'
                                  if venv is not None else None)

    def get_venv_activation_line(self) -> str:
        if self.activation_script:
            return (f'source {self.activation_script}\n'
                    f'echo "venv: {self.activation_script}"\n')
        return ''

    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        """Submit a task."""
        raise NotImplementedError

    def cancel(self, id: int) -> None:
        """Cancel a task."""
        raise NotImplementedError

    def get_ids(self) -> set[int]:
        """Get ids for all tasks the scheduler knows about."""
        raise NotImplementedError

    def hold(self, id: int) -> None:
        raise NotImplementedError

    def release_hold(self, id: int) -> None:
        raise NotImplementedError

    def error_file(self, task: Task) -> Path:
        return task.folder / f'{task.cmd.short_name}.{task.id}.err'

    def has_timed_out(self, task: Task) -> bool:
        path = self.error_file(task).expanduser()
        if path.is_file():
            task.tstop = path.stat().st_mtime
            lines = path.read_text().splitlines()
            for line in lines:
                if line.endswith('DUE TO TIME LIMIT ***'):
                    return True
        return False

    def maxrss(self, id: int) -> int:
        return 0

    def get_config(self, queue: str = '') -> tuple[list[tuple[str, int, str]],
                                                   list[str]]:
        raise NotImplementedError



from __future__ import annotations
from myqueue.workflow import run, wrap, resources


def f(x: int) -> int:
    print(x)
    return x + 1


def work():
    A = []
    for x in range(3):
        y = f(x)
        f(y + 1)
        A.append(y)

    b = max(*A)

    if b > 2:
        print(b)

    return b


@resources(tmax='1h')
def workflow():
    A = []
    for x in range(3):
        y = wrap(f, name=f'fa-{x}')(x)
        run(function=f, name=f'fb-{x}', args=[y + 1])
        A.append(y)

    b = wrap(max)(*A)

    if b > 2:
        with resources(tmax='10s'):
            wrap(print)(b)

    return b


if __name__ == '__main__':
    workflow()



from __future__ import annotations
import os
from myqueue.utils import convert_done_files


def test_convert_done_files(tmp_path):
    (tmp_path / 't1.done').write_text('')
    (tmp_path / 't2.done').write_text('[117]')
    os.chdir(tmp_path)
    convert_done_files()
    t1 = (tmp_path / 't1.state').read_text()
    t2 = (tmp_path / 't2.state').read_text()
    assert t1 == '{"state": "done"}\n'
    assert t2 == '{"state": "done",\n "result": [117]}\n'



from __future__ import annotations
import os
import sys
from collections import defaultdict
from pathlib import Path

from myqueue.task import Task
from myqueue.utils import mqhome


def colored(state: str) -> str:
    """Yellow for running, red for bad, green for done."""
    if state.isupper():
        return '\033[91m' + state + '\033[0m'
    if state.startswith('done'):
        return '\033[92m' + state + '\033[0m'
    if state.startswith('running'):
        return '\033[93m' + state + '\033[0m'
    return state


def pprint(tasks: list[Task],
           *,
           verbosity: int = 1,
           columns: str = 'ifnaIrAste',
           short: bool = False,
           maxlines: int = 9999999999,
           sort: str | None = None,
           reverse: bool = False) -> None:
    """Pretty-print tasks.

    Use short=True to get only a summary.
    """
    if verbosity < 0:
        return

    if not tasks:
        return

    if isinstance(sort, str):
        tasks = sorted(tasks,
                       key=lambda task: task.order_key(sort),  # type: ignore
                       reverse=reverse)

    home = str(mqhome()) + '/'
    cwd = str(Path.cwd()) + '/'

    if columns.endswith('-'):
        columns = ''.join(c for c in 'ifnaIrAste' if c not in columns[:-1])

    titles = ['id', 'folder', 'name', 'args', 'info',
              'res.', 'age', 'state', 'time', 'error']
    c2i = {c: i for i, c in enumerate('ifnaIrAste')}
    indices = [c2i[c] for c in columns]

    if len(tasks) > maxlines:
        cut1 = maxlines // 2
        cut2 = maxlines - cut1 - 2
        skipped = len(tasks) - cut1 - cut2
        tasks = tasks[:cut1] + tasks[-cut2:]
    else:
        skipped = 0

    lines = []
    lengths = [0] * len(columns)

    count: dict[str, int] = defaultdict(int)
    for task in tasks:
        words = task.words()
        _, folder, _, _, _, _, _, state, _, _ = words
        count[state] += 1
        if folder.startswith(cwd):
            words[1] = './' + folder[len(cwd):]
        elif folder.startswith(home):
            words[1] = '~/' + folder[len(home):]
        words = [words[i] for i in indices]
        lines.append(words)
        lengths = [max(n, len(word)) for n, word in zip(lengths, words)]

    # remove empty columns ...
    lines = [[word for word, length in zip(words, lengths) if length]
             for words in lines]
    columns = ''.join(c for c, length in zip(columns, lengths) if length)
    lengths = [length for length in lengths if length]

    if skipped:
        lines[cut1:cut1] = [[f'... ({skipped} tasks not shown)']]

    if verbosity:
        lines[:0] = [[titles[c2i[c]] for c in columns]]
        lengths = [max(length, len(title))
                   for length, title in zip(lengths, lines[0])]

    try:
        N = os.get_terminal_size().columns
        cut = max(0, N - sum(L + 1 for L, c in zip(lengths, columns)
                             if c != 'e'))
    except OSError:
        cut = 999999

    use_color = sys.stdout.isatty() and 'MYQUEUE_TESTING' not in os.environ

    if verbosity:
        lines[1:1] = [['─' * L for L in lengths]]
        lines.append(lines[1])

    if not short:
        for i, words in enumerate(lines):
            words2 = []
            for word, c, L in zip(words, columns, lengths):
                if c == 'e':
                    word = word[:cut]
                elif c in 'At':
                    word = word.rjust(L)
                else:
                    word = word.ljust(L)
                    if c == 's' and use_color:
                        word = colored(word)
                words2.append(word)
            line = ' '.join(words2)
            if use_color:
                if i == 0:
                    line = '\033[94m' + line + '\033[0m'
                elif (i == 1 or i == len(lines) - 1) or not verbosity:
                    line = '\033[93m' + line + '\033[0m'
            print(line)

    if verbosity:
        count['total'] = len(tasks)
        print(', '.join(f'{colored(state) if use_color else state}: {n}'
                        for state, n in count.items()))



from __future__ import annotations

import shutil
import time
from pathlib import Path

import pytest
from myqueue.queue import Queue, sort_out_dependencies
from myqueue.task import task
from myqueue.states import State
from myqueue.utils import chdir, get_states_of_active_tasks
from myqueue.cli import _main

LOCAL = True


def test_submit_args(mq):
    mq('''submit "myqueue.test.mod --opt={...,x={'y':(1,2)}"''')
    mq.wait()
    assert mq.states() == 'd'


def test_submit(mq):
    f = Path('folder')
    f.mkdir()
    mq('submit time@sleep+0.1 . folder --max-tasks=9')
    mq('submit shell:echo+hello -d time@sleep+0.1')
    assert mq.wait() == 'ddd'
    shutil.rmtree(f)
    mq('sync -z')
    mq('sync')
    assert mq.states() == 'dd'
    mq('daemon status')
    assert get_states_of_active_tasks() == {}


def test_fail(mq):
    mq('submit time@sleep+a')
    mq('submit shell:echo+hello -d time@sleep+a')
    mq('submit shell:echo+hello2 -d shell:echo+hello')
    mq.wait()
    mq('info')
    mq('info -i1 -v')
    mq('ls -S t')
    assert mq.states() == 'FCC', mq.states()
    mq('resubmit -sF . -z')
    assert mq.states() == 'FCC'
    mq('resubmit -sF . --remove')
    assert mq.wait() == 'CCF'


def test_fail2(mq):
    mq('submit time@sleep+a --workflow')
    mq.wait()
    assert mq.states() == 'F'
    mq('remove --states F --force .')
    mq('submit time@sleep+a --workflow')
    assert mq.wait() == 'F'


def test_timeout(mq):
    t = 3 if LOCAL else 120
    mq(f'submit -n zzz "shell:sleep {t}" -R 1:1s')
    mq('submit "shell:echo hello" -d zzz')
    mq.wait()
    mq('resubmit -sT . -R 1:5m --remove')
    assert mq.wait() == 'Cd'


def test_timeout2(mq):
    t = 3 if LOCAL else 120
    mq(f'submit "shell:sleep {t}" -R1:{t // 3}s --restart 3')
    mq(f'submit "shell:echo hello" -d shell:sleep+{t}')
    mq.wait()
    mq('ls')
    mq('kick')
    mq('ls')
    mq.wait()
    mq('kick')
    q = mq.wait()
    if q == 'TCTCTC':
        # on slow systems we need another kick:
        mq('kick')
        q = mq.wait()[2:]
    assert q == 'TCTCdd'


def test_oom(mq):
    mq(f'submit "myqueue.test@oom {LOCAL}" --restart 2')
    assert mq.wait() == 'M'
    mq('kick')
    assert mq.wait() == 'Md'


def test_cancel(mq):
    mq('submit shell:sleep+2')
    mq('submit shell:sleep+999')
    mq('submit shell:echo+hello -d shell:sleep+999')
    mq('rm -n shell:sleep+999 -srq .')
    assert mq.wait() == 'dC'


def test_check_dependency_order(mq):
    mq('submit myqueue.test@timeout_once -R 1:5s --restart 1')
    mq('submit shell:echo+ok -d myqueue.test@timeout_once --restart 1')
    assert mq.wait() == 'TC'
    mq('kick -z')
    mq('kick')
    assert mq.wait() == 'TCdd'


def test_misc(mq):
    f = Path('subfolder')
    f.mkdir()
    with chdir(f):
        mq('init')
        mq('init')
    mq('help')
    mq('ls -saA')
    mq('-V')
    mq('completion')
    mq('completion -v')
    mq('ls no_such_folder', error=1)
    mq('')
    mq('info -A')


def test_sync_kick(mq):
    mq('sync')
    mq('kick')


def test_slash(mq):
    mq('submit "shell:echo a/b"')
    mq('submit "shell:echo a/c" -w')
    assert mq.wait() == 'dd'


def test_config(mq):
    mq('config local')


def test_more_homes(mq):
    f = Path('folder')
    f.mkdir()
    with chdir(f):
        mq('init')
    mq('submit shell:echo . folder', error=1)


def test_permission_error(mq):
    mq('ls')
    try:
        (mq.config.home / '.myqueue').chmod(0o500)  # r-x
        mq('ls')
    finally:
        (mq.config.home / '.myqueue').chmod(0o700)  # rwx


def test_failing_scheduler(mq):
    with pytest.raises(RuntimeError):
        # Special argument that makes test-scheduler raise an error:
        mq('submit "time.sleep FAIL"')
    assert mq.wait() == ''


@pytest.mark.xfail
def test_ctrl_c(mq):
    # Special argument that makes test-scheduler raise an error:
    mq('submit "time.sleep SIMULATE-CTRL-C"')
    assert mq.wait() == 'd'


def test_sync_cancel(mq):
    with Queue(mq.config) as q:
        t = task('shell:echo')
        t.state = State.running
        t.trunning = time.time()
        q.add(t)
    mq('sync')
    assert mq.states() == ''


def test_hold_release(mq):
    mq('submit shell:echo+hello')
    mq('modify -s q -N h . -z')
    mq('modify -s q -N h .')
    assert mq.wait() == 'h'
    (name, state), = get_states_of_active_tasks().items()
    assert state == 'h'
    assert name == './shell:echo+hello'

    mq('modify -s h -N q . -z')
    mq('modify -s h -N q .')
    assert mq.wait() == 'd'
    with pytest.raises(ValueError):
        mq('modify -s d -N q .')


def test_clean_up(mq):
    with Queue(mq.config) as q:
        t1 = task('shell:echo+1')
        t1.id = 1
        t1.state = State.running
        t1.trunning = 0.0  # very old
        t2 = task('shell:echo+2', deps=[t1])
        t2.state = State.queued
        t2.id = 2
        sort_out_dependencies([t1, t2], q)
        q.add(t1, t2)
    assert mq.states() == 'TC'


def test_cli_exception(mq, monkeypatch):
    def run(args, test):
        raise ValueError

    monkeypatch.setattr('myqueue.cli.run', run)

    # With --traceback:
    with pytest.raises(ValueError):
        mq('ls')

    # Without --traceback:
    assert _main(['ls']) == 1


def test_mq_exception(mq):
    mq('rm', error=1)
    mq('ls -i 0')
    mq('ls -i 0 -s q', error=1)
    with pytest.raises(ValueError):
        mq('ls -i 0 .')
    mq('rm .', error=1)



from __future__ import annotations
import io
import os
import re
import sys
from pathlib import Path
from subprocess import PIPE, run

import pytest

user = os.environ.get('USER', 'root')
docs = Path(__file__).parent / '../../docs'


def skip(line: str) -> bool:
    for s in ['<task>', '<states>', 'ABC 123']:
        if s in line:
            return True
    return False


def run_document(mq, path: Path, test=False, update=False) -> None:
    assert path.is_file()
    lines = path.read_text().splitlines()
    blocks: list[tuple[str, list[str], int]] = []
    n = 0
    while n < len(lines):
        line = lines[n]
        if line.endswith('::'):
            line = lines[n + 2]
            if line[:5] == '    $' and not skip(line):
                cmd = ''
                output: list[str] = []
                L = 0
                for n, line in enumerate(lines[n + 2:], n + 2):
                    if not line:
                        break
                    if line[4] == '$':
                        if cmd:
                            blocks.append((cmd, output, L))
                        cmd = line[6:]
                        output = []
                        L = n
                    else:
                        output.append(line)
                blocks.append((cmd, output, L))
        n += 1

    pypath = Path().absolute()
    offset = 0
    folder = '.'
    errors = 0
    for cmd, output, L in blocks:
        print('$', cmd)
        if cmd.startswith('mq '):
            os.chdir(folder)
            out = sys.stdout
            sys.stdout = io.StringIO()
            try:
                mq(cmd[3:])
                mq.wait()
            finally:
                actual_output = sys.stdout.getvalue().splitlines()[1:]
                sys.stdout = out
        else:
            actual_output, folder = run_command(cmd, folder, pypath)
        actual_output = ['    ' + line.rstrip()
                         for line in actual_output]
        errors += compare(output, actual_output)
        L += 1 + offset
        lines[L:L + len(output)] = actual_output
        offset += len(actual_output) - len(output)

    if update:
        path.write_text('\n'.join(lines) + '\n')

    if test:
        assert errors == 0


def run_command(cmd: str,
                folder: str,
                pypath: Path) -> tuple[list[str], str]:
    if cmd.startswith('#'):
        return [], folder
    cmd, _, _ = cmd.partition('  #')
    env = os.environ.copy()
    env['PYTHONPATH'] = str(pypath)
    env['LC_ALL'] = 'C'
    result = run(f'cd {folder}; {cmd}; pwd',
                 shell=True, check=True, stdout=PIPE, env=env)
    output = result.stdout.decode().splitlines()
    folder = output.pop()
    return output, folder


def clean(line):
    line = re.sub(r'[A-Z]?[a-z]+\s+[0-9]+\s+[0-9]+:[0-9]+', '############',
                  line)
    line = re.sub(r' 0:[0-9][0-9]', ' 0:##', line)
    line = re.sub(r'[rw.-]{10,11}', '##########', line)
    line = re.sub(r' tot\w+ \d+', ' ##### #', line)
    line = re.sub(rf' {user} \w+ ', ' jensj ##### ', line)
    line = re.sub(r' jensj jensj ', ' jensj ##### ', line)
    return line


def compare(t1, t2):
    t1 = [clean(line) for line in t1]
    t2 = [clean(line) for line in t2]
    if t1 == t2 or '    ...' in t1:
        return 0
    print('<<<<<<<<<<<')
    print('\n'.join(t1))
    print('===========')
    print('\n'.join(t2))
    print('>>>>>>>>>>>')
    return 1


@pytest.mark.docs
def test_docs_workflows(mq, monkeypatch):
    monkeypatch.syspath_prepend('.')
    p = Path('prime')
    p.mkdir()
    for f in docs.glob('prime/*.*'):
        (p / f.name).write_text(f.read_text())
    venv = Path('venv').absolute()
    os.environ['VIRTUAL_ENV'] = str(venv)
    venv.mkdir()
    bin = venv / 'bin'
    bin.mkdir()
    (bin / 'activate').write_text(
        f'export PYTHONPATH={venv.parent}\n')
    mq.scheduler.activation_script = bin / 'activate'
    run_document(mq, docs / 'workflows.rst', test=True)


@pytest.mark.docs
def test_docs_documentation(mq):
    run_document(mq, docs / 'documentation.rst', test=True)


@pytest.mark.docs
def test_docs_quickstart(mq):
    run_document(mq, docs / 'quickstart.rst', test=True)



from __future__ import annotations

import subprocess
import sys

from myqueue.config import Configuration
from myqueue.schedulers import Scheduler
from myqueue.states import State
from myqueue.task import Task


class TestScheduler(Scheduler):
    current_scheduler: TestScheduler | None = None

    def __init__(self, config: Configuration):
        Scheduler.__init__(self, config)
        self.folder = self.config.home / '.myqueue'
        self.tasks: list[Task] = []
        self.number = 0

    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        if dry_run:
            return 1
        if task.cmd.args == ['FAIL']:
            raise RuntimeError
        if task.dtasks:
            ids = {t.id for t in self.tasks}
            for t in task.dtasks:
                assert t.id in ids
        self.number += 1
        task.state = State.queued
        self.tasks.append(task)
        return self.number

    def cancel(self, id: int) -> None:
        for i, task in enumerate(self.tasks):
            if task.id == id:
                break
        else:
            return
        del self.tasks[i]

    def hold(self, id: int) -> None:
        for task in self.tasks:
            if task.id == id:
                break
        else:
            raise ValueError('No such task!')
        task.state = State.hold

    def release_hold(self, id: int) -> None:
        for task in self.tasks:
            if task.id == id:
                break
        else:
            raise ValueError('No such task!')
        task.state = State.queued

    def get_ids(self) -> set[int]:
        return {task.id for task in self.tasks}

    def kick(self) -> bool:
        for task in self.tasks:
            if task.state == 'queued' and not task.deps:
                break
        else:
            return True

        self.run(task)
        return False

    def run(self, task: Task) -> None:
        out = f'{task.cmd.short_name}.{task.id}.out'
        err = f'{task.cmd.short_name}.{task.id}.err'

        cmd = str(task.cmd)
        cmd = cmd.replace('python3', sys.executable)
        if task.resources.processes > 1:
            n = task.resources.processes
            cmd = f'MYQUEUE_TEST_NPROCESSES={n} ' + cmd
        cmd = f'cd {task.folder} && {cmd} 2> {err} > {out}'
        activation_script = self.activation_script
        if activation_script:
            cmd = f'. {activation_script} && ' + cmd
        (self.folder / f'test-{task.id}-0').write_text('')
        tmax = task.resources.tmax
        try:
            result = subprocess.run(cmd,
                                    shell=True,
                                    check=not True,
                                    timeout=tmax)
        except subprocess.TimeoutExpired:
            state = State.TIMEOUT
        else:
            if result.returncode == 0:
                state = State.done
            else:
                state = State.FAILED
        self.update(task, state)

    def update(self, task: Task, state: State) -> None:
        n = {State.done: 1,
             State.FAILED: 2,
             State.TIMEOUT: 3}[state]

        (self.folder / f'test-{task.id}-{n}').write_text('')

        if state == 'done':
            tasks = []
            for j in self.tasks:
                if j is not task:
                    if task.dname in j.deps:
                        j.deps.remove(task.dname)
                    tasks.append(j)
            self.tasks = tasks
        else:
            task.state = State.CANCELED
            self.cancel_dependents(task)
            self.tasks = [task for task in self.tasks
                          if task.state != 'CANCELED']

    def cancel_dependents(self, task: Task) -> None:
        for job in self.tasks:
            if task.dname in job.deps:
                job.state = State.CANCELED
                self.cancel_dependents(job)



import json
import os

from myqueue.queue import Queue, dump_db
from myqueue.task import create_task


def test_migration(tmp_path):
    os.chdir(tmp_path)
    mq = tmp_path / '.myqueue'
    mq.mkdir()
    task = create_task('time@sleep')
    dct = {'tasks': [task.todict()]}
    text = json.dumps(dct)
    (mq / 'queue.json').write_text(text)
    with Queue() as q:
        tasks = q.select()
        assert len(tasks) == 1
    dump_db(mq / 'queue.sqlite3')



"""Module for testing a module as a task."""
from __future__ import annotations
import sys


def main():
    """Print CLI-args."""
    for arg in sys.argv:
        print(arg)


if __name__ == '__main__':
    main()



from __future__ import annotations
import pytest
from myqueue.config import Configuration


def test_config():
    cfg = Configuration('test')
    print(cfg)
    print(repr(cfg))


def test_missing_or_deprecated_key(mq):
    cfg_file = mq.config.home / '.myqueue/config.py'
    cfg_file.write_text('config = {}\n')
    with pytest.raises(ValueError):
        Configuration.read()
    cfg_file.write_text("config = {'scheduler': 'test', 'mpi': 'openmpi'}\n")
    with pytest.warns(UserWarning):
        Configuration.read()



from __future__ import annotations
from myqueue.queue import Queue
from math import inf


def hold_or_release(queue: Queue) -> dict[str, int]:
    maxweight = queue.config.maximum_total_task_weight
    if maxweight == inf:
        return {}
    totweight = 0
    queued = []
    held = []
    sql = (
        'SELECT id, state, weight FROM tasks '
        'WHERE state IN ("q", "r", "h", "F", "M", "T") AND '
        'weight != 0 AND user = ?')
    for id, state, weight in queue.sql(sql, [queue.config.user]):
        if state in 'qrFMT':
            totweight += weight
        if state == 'q':
            queued.append((id, weight))
        elif state == 'h':
            held.append((id, weight))

    changes: list[tuple[str, int]] = []

    if totweight > maxweight:
        for id, weight in queued:
            queue.scheduler.hold(id)
            changes.append(('h', id))
            totweight -= weight
            if totweight < maxweight:
                break
    elif totweight < maxweight:
        for id, weight in held[::-1]:
            queue.scheduler.release_hold(id)
            changes.append(('q', id))
            totweight += weight
            if totweight > maxweight:
                break

    if not changes:
        return {}

    with queue.connection as con:
        con.executemany(
            'UPDATE tasks SET state = ? WHERE id = ?', changes)

    return {changes[0][0]: len(changes)}



#!/usr/bin/env python3
"""Bash completion.

Put this in your .bashrc::

    complete -o default -C "python3 -m myqueue.complete" mq

"""

from __future__ import annotations
import os
import sys
from typing import Any, Iterable, Mapping


def read() -> dict[str, Any]:
    """Read queue as a dict."""
    from pathlib import Path
    import json
    from myqueue.config import find_home_folder
    home = find_home_folder(Path('.').resolve())
    path = home / '.myqueue/queue.json'
    try:
        dct: dict[str, Any] = json.loads(path.read_text())
        return dct
    except Exception:
        return {}


# Beginning of computer generated data:
commands = {
    'completion':
        ['-v', '--verbose', '-q', '--quiet', '-T', '--traceback'],
    'config':
        ['-Q', '--queue-name', '--in-place', '-z', '--dry-run', '-v',
         '--verbose', '-q', '--quiet', '-T', '--traceback'],
    'daemon':
        ['-z', '--dry-run', '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback'],
    'help':
        [''],
    'info':
        ['-v', '--verbose', '-q', '--quiet', '-T', '--traceback', '-i',
         '--id', '-A', '--all'],
    'init':
        ['-z', '--dry-run', '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback'],
    'kick':
        ['-z', '--dry-run', '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback'],
    'list':
        ['-s', '--states', '-i', '--id', '-n', '--name', '-e', '--error',
         '-c', '--columns', '-S', '--sort', '-C', '--count',
         '--not-recursive', '-v', '--verbose', '-q', '--quiet',
         '-T', '--traceback'],
    'modify':
        ['-E', '--email', '-N', '--new-state', '-s', '--states', '-i',
         '--id', '-n', '--name', '-e', '--error', '-z',
         '--dry-run', '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback', '-r', '--recursive'],
    'remove':
        ['-f', '--force', '-s', '--states', '-i', '--id', '-n', '--name',
         '-e', '--error', '-z', '--dry-run', '-v', '--verbose',
         '-q', '--quiet', '-T', '--traceback', '-r',
         '--recursive'],
    'resubmit':
        ['--remove', '-R', '--resources', '-w', '--workflow', '-s',
         '--states', '-i', '--id', '-n', '--name', '-e',
         '--error', '-z', '--dry-run', '-v', '--verbose', '-q',
         '--quiet', '-T', '--traceback', '-r', '--recursive'],
    'submit':
        ['-d', '--dependencies', '-n', '--name', '--restart',
         '--max-tasks', '-R', '--resources', '-w', '--workflow',
         '-z', '--dry-run', '-v', '--verbose', '-q', '--quiet',
         '-T', '--traceback'],
    'sync':
        ['-z', '--dry-run', '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback'],
    'workflow':
        ['--max-tasks', '-f', '--force', '-t', '--targets', '-p',
         '--pattern', '-a', '--arguments', '-z', '--dry-run',
         '-v', '--verbose', '-q', '--quiet', '-T',
         '--traceback']}
# End of computer generated data

aliases = {'rm': 'remove',
           'ls': 'list'}


def complete(word: str, previous: str, line: str, point: int) -> Iterable[str]:
    for w in line[:point - len(word)].strip().split()[1:]:
        if w[0].isalpha():
            if w in commands or w in aliases:
                command = aliases.get(w, w)
                break
    else:
        opts = ['-h', '--help', '-V', '--version']
        if word[:1] == '-':
            return opts
        return list(commands) + list(aliases) + opts

    if word[:1] == '-':
        return commands[command]

    if previous in ['-n', '--name']:
        dct = read()
        words = set()
        for task in dct['tasks']:
            cmd = task['cmd']
            words.add((cmd['cmd'] + '+' + '_'.join(cmd['args'])).rstrip('+'))
        return words

    if previous in ['-i', '--id']:
        dct = read()
        return {str(task['id']) for task in dct['tasks']}

    if command == 'help':
        return [cmd for cmd in (list(commands) + list(aliases))
                if cmd != 'help']

    if command == 'daemon':
        return ['start', 'stop', 'status']

    return []


def main(environ: Mapping[str, str], word: str, previous: str) -> None:
    line = environ['COMP_LINE']
    point = int(environ['COMP_POINT'])
    words = complete(word, previous, line, point)
    for w in words:
        if w.startswith(word):
            print(w)


if __name__ == '__main__':
    word, previous = sys.argv[2:]
    main(os.environ, word, previous)  # pragma: no cover



from __future__ import annotations
import os
import sys
import time
from pathlib import Path


def oom(local: bool = True) -> None:
    n = int(os.environ.get('MYQUEUE_TEST_NPROCESSES', '1'))
    if n == 1:
        print('error: exceeded memory limit at some point.', file=sys.stderr)
        raise MemoryError


def timeout_once():
    path = Path('timeout_once.out')
    if not path.is_file():
        path.touch()
        time.sleep(10)



from __future__ import annotations

import time
from pathlib import Path
from types import TracebackType
from typing import Sequence, TypeVar, TYPE_CHECKING

from myqueue.pretty import pprint
from myqueue.queue import Queue, sort_out_dependencies
from myqueue.schedulers import Scheduler
from myqueue.states import State
from myqueue.task import Task
from myqueue.utils import plural

if TYPE_CHECKING:
    import rich.progress as progress

TaskName = Path


def submit(queue: Queue,
           tasks: Sequence[Task],
           *,
           done: list[Task] = None,
           max_tasks: int = 1_000_000_000,
           verbosity: int = 1) -> None:
    """Submit tasks to queue.

    Parameters
    ==========
    force: bool
        Ignore and remove name.FAILED files.
    """

    sort_out_dependencies(tasks, queue, done)

    tasks = [task for task in order({task: task.dtasks for task in tasks})
             if task.state == State.undefined]

    tasks = tasks[:max_tasks]

    default_weight = queue.config.default_task_weight
    for task in tasks:
        task.resources.set_default_weight(default_weight)

    ids, ex = submit_tasks(queue.scheduler, tasks, verbosity, queue.dry_run)
    submitted = tasks[:len(ids)]

    if ex:
        nskip = len(tasks) - len(submitted)
        print()
        print('Skipped', plural(nskip, 'task'))

    t = time.time()
    for task, id in zip(submitted, ids):
        task.id = id
        task.state = State.queued
        task.tqueued = t

    pprint(submitted, verbosity=0, columns='ifnaIr',
           maxlines=10 if verbosity < 2 else 99999999999999)
    if submitted:
        if queue.dry_run:
            print(plural(len(submitted), 'task'), 'to submit')
        else:
            queue.add(*submitted)
            print(plural(len(submitted), 'task'), 'submitted')

    if ex:
        raise ex


def submit_tasks(scheduler: Scheduler,
                 tasks: Sequence[Task],
                 verbosity: int,
                 dry_run: bool) -> tuple[list[int],
                                         Exception | KeyboardInterrupt | None]:
    """Submit tasks."""
    import rich.progress as progress

    ids = []
    ex = None

    pb: progress.Progress | NoProgressBar

    if verbosity and len(tasks) > 1:
        pb = progress.Progress('[progress.description]{task.description}',
                               progress.BarColumn(),
                               progress.MofNCompleteColumn())
    else:
        pb = NoProgressBar()

    with pb:
        try:
            pid = pb.add_task('Submitting tasks:', total=len(tasks))
            for task in tasks:
                id = scheduler.submit(
                    task,
                    dry_run,
                    verbosity >= 2)
                ids.append(id)
                task.id = id
                pb.advance(pid)
        except (Exception, KeyboardInterrupt) as x:
            ex = x

    return ids, ex


T = TypeVar('T')


def order(nodes: dict[T, list[T]]) -> list[T]:
    """Depth first.

    >>> order({1: [2], 2: [], 3: [4], 4: []})
    [2, 1, 4, 3]
    """
    import networkx as nx  # type: ignore
    result: list[T] = []
    g = nx.Graph(nodes)
    for component in nx.connected_components(g):
        dg = nx.DiGraph({node: nodes[node]
                         for node in component
                         if node in nodes})
        result += reversed(list(nx.topological_sort(dg)))
    return result


class NoProgressBar:
    """Dummy progress-bar."""
    def __enter__(self) -> NoProgressBar:
        return self

    def __exit__(self,
                 type: Exception,
                 value: Exception,
                 tb: TracebackType) -> None:
        pass

    def add_task(self, text: str, total: int) -> progress.TaskID:
        import rich.progress as progress
        return progress.TaskID(0)

    def advance(self, id: progress.TaskID) -> None:
        pass



from __future__ import annotations

from myqueue.email import configure_email
from myqueue.queue import Queue
from myqueue.selection import Selection
from myqueue.states import State


def modify(queue: Queue,
           selection: Selection,
           newstate: State,
           email: set[State]) -> None:
    """Modify task(s)."""
    tasks = queue.select(selection)

    if email != {State.undefined}:
        configure_email(queue.config)
        if queue.dry_run:
            print(tasks, email)
        else:
            n = ''.join(state.value for state in email)
            with queue.connection as con:
                con.executemany(
                    f'UPDATE tasks SET notifications = "{n}" WHERE id = ?',
                    [(task.id,) for task in tasks])

    if newstate == State.undefined:
        return

    if newstate == State.queued:
        oldstate = State.hold
        operation = queue.scheduler.release_hold
    else:
        assert newstate == State.hold
        oldstate = State.queued
        operation = queue.scheduler.hold

    if any(task.state != oldstate for task in tasks):
        raise ValueError(f'Initial state must be: {oldstate}!')

    if not queue.dry_run:
        for task in tasks:
            operation(task.id)

        with queue.connection as con:
            con.executemany(
                f'UPDATE tasks SET state = "{newstate.value}" WHERE id = ?',
                [(task.id,) for task in tasks])

    print(f'{oldstate} -> {newstate}: {len(tasks)}')



import json
from pathlib import Path
from sqlite3 import Connection

from myqueue.task import Task


def migrate(jsonfile: Path, con: Connection) -> None:
    print(f'Converting {jsonfile} to SQLite3 file ...',
          end='', flush=True)
    text = jsonfile.read_text()
    data = json.loads(text)
    root = jsonfile.parent.parent
    ids = {}
    tasks = []
    for dct in data['tasks']:
        task = Task.fromdict(dct, root)
        ids[task.dname] = task.id
        tasks.append(task)
    deps = []
    for task in tasks:
        for dep in task.deps:
            deps.append((task.id, ids[dep]))
    with con:
        q = ', '.join('?' * 17)
        con.executemany(
            f'INSERT INTO tasks VALUES ({q})',
            [task.to_sql(root) for task in tasks])
        con.executemany(
            'INSERT INTO dependencies VALUES (?, ?)', deps)
    jsonfile.with_suffix('.old.json').write_text(text)
    jsonfile.unlink()
    print(' done')



from __future__ import annotations
import json
import os

from myqueue.complete import complete, main


def test_ls():
    words = complete('-', 'ls', 'mq ls -', 7)
    assert '--not-recursive' in words


def test_ls_main(capsys):
    main({'COMP_LINE': 'mq ls -', 'COMP_POINT': '7'}, '-', 'ls')
    out = capsys.readouterr().out
    words = out.splitlines()
    assert '--not-recursive' in words


def test_daemon():
    words = complete('', 'daemon', 'mq daemon ', 9)
    assert 'start' in words


def test_rm():
    words = complete('', 'mq', 'mq ', 3)
    assert 'rm' in words


def test_help():
    words = complete('', 'help', 'mq help ', 8)
    assert 'rm' in words
    assert 'remove' in words


def test_bare():
    words = complete('-', 'mq', 'mq -', 4)
    assert '-V' in words


def test_sync():
    words = complete('', 'sync', 'mq sync ', 8)
    assert words == []


def test_read(tmp_path):
    os.chdir(tmp_path)
    mq = tmp_path / '.myqueue'
    mq.mkdir()
    dct = {'tasks': [{'cmd': {'cmd': 'abc123', 'args': []}, 'id': 117}]}
    (mq / 'queue.json').write_text(json.dumps(dct))
    words = complete('', '-n', 'mq ls -n ', 9)
    assert words == {'abc123'}
    words = complete('', '-i', 'mq ls -i ', 9)
    assert words == {'117'}



from __future__ import annotations
import os
import subprocess
from pathlib import Path

from myqueue.task import Task
from myqueue.schedulers import Scheduler, SchedulerError


class PBS(Scheduler):
    output_file_pattern = r'[a-zA-Z]+.*\.[eo]+[1-9]+[0-9]*'

    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        nodelist = self.config.nodes
        nodes, nodename, nodedct = task.resources.select(nodelist)

        processes = task.resources.processes

        if processes < nodedct['cores']:
            ppn = processes
        else:
            assert processes % nodes == 0
            ppn = processes // nodes

        hours, rest = divmod(task.resources.tmax, 3600)
        minutes, seconds = divmod(rest, 60)

        qsub = ['qsub',
                '-N',
                task.cmd.short_name,
                '-l',
                f'walltime={hours}:{minutes:02}:{seconds:02}',
                '-l',
                f'nodes={nodes}:ppn={ppn}',
                '-d', str(task.folder)]

        qsub += self.config.extra_args + nodedct.get('extra_args', [])

        if task.dtasks:
            ids = ':'.join(str(tsk.id) for tsk in task.dtasks)
            qsub.extend(['-W', f'depend=afterok:{ids}'])

        cmd = str(task.cmd)
        if task.resources.processes > 1:
            mpiexec = 'mpiexec -x OMP_NUM_THREADS=1 -x MPLBACKEND=Agg '
            if 'mpiargs' in nodedct:
                mpiexec += nodedct['mpiargs'] + ' '
            cmd = mpiexec + cmd.replace('python3',
                                        self.config.parallel_python)
        else:
            cmd = 'MPLBACKEND=Agg ' + cmd

        home = self.config.home

        script = '#!/bin/bash -l\n'

        script += self.get_venv_activation_line()

        script += (
            '#!/bin/bash -l\n'
            'id=${PBS_JOBID%.*}\n'
            f'mq={home}/.myqueue/pbs-$id\n'
            f'(touch $mq-0 && cd {task.folder} && {cmd} && touch $mq-1) || '
            'touch $mq-2\n')

        if dry_run:
            print(qsub, script)
            return 1

        p = subprocess.run(qsub,
                           input=script.encode(),
                           capture_output=True)
        if p.returncode:
            raise SchedulerError((p.stderr + p.stdout).decode())
        id = int(p.stdout.split(b'.')[0].decode())
        return id

    def error_file(self, task: Task) -> Path:
        return task.folder / f'{task.cmd.short_name}.e{task.id}'

    def cancel(self, id: int) -> None:
        subprocess.run(['qdel', str(id)])

    def get_ids(self) -> set[int]:
        user = os.environ.get('USER', 'test')
        cmd = ['qstat', '-u', user]
        p = subprocess.run(cmd, stdout=subprocess.PIPE)
        queued = {int(line.split()[0].split(b'.')[0].decode())
                  for line in p.stdout.splitlines()
                  if line[:1].isdigit()}
        return queued
